import os
import shutil
from pathlib import Path
from typing import TypeVar, Type, Tuple

from alembic import script, command
from alembic.config import Config
from alembic.runtime import migration
from sqlalchemy import create_engine
import sqlalchemy.engine
from sqlalchemy.orm import sessionmaker

from entropylab.api.errors import EntropyError
from entropylab.logger import logger
from entropylab.results_backend.sqlalchemy.model import Base, ResultTable, MetadataTable
from entropylab.results_backend.sqlalchemy.project import project_name, project_path
from entropylab.results_backend.sqlalchemy.storage import HDF5Storage, EntityType

T = TypeVar("T", bound=Base)

_SQL_ALCHEMY_MEMORY = ":memory:"
_ENTROPY_DIRNAME = ".entropy"
_DB_FILENAME = "entropy.db"
_HDF5_FILENAME = "entropy.hdf5"


class _DbInitializer:
    def __init__(self, path: str, echo=False):
        """
        :param path: path to directory containing Entropy project
        :param echo: if True, the database engine will log all statements
        """
        self._validate_path(path)

        if path is None or path == _SQL_ALCHEMY_MEMORY:
            logger.debug("_DbInitializer is in in-memory mode")
            self._storage = HDF5Storage()
            self._engine = create_engine("sqlite:///" + _SQL_ALCHEMY_MEMORY, echo=echo)
            self._alembic_util = _AlembicUtil(self._engine)
        else:
            logger.debug("_DbInitializer is in project directory mode")
            creating_new = not os.path.isdir(path)
            entropy_dir_path = os.path.join(path, _ENTROPY_DIRNAME)
            os.makedirs(entropy_dir_path, exist_ok=True)
            logger.debug(f"Entropy directory is at: {entropy_dir_path}")

            db_file_path = os.path.join(entropy_dir_path, _DB_FILENAME)
            logger.debug(f"DB file is at: {db_file_path}")

            hdf5_file_path = os.path.join(entropy_dir_path, _HDF5_FILENAME)
            logger.debug(f"hdf5 file is at: {hdf5_file_path}")

            self._engine = create_engine("sqlite:///" + db_file_path, echo=echo)
            self._storage = HDF5Storage(hdf5_file_path)
            self._alembic_util = _AlembicUtil(self._engine)
            if creating_new:
                self._print_project_created(path)

    def init_db(self) -> Tuple[sqlalchemy.engine.Engine, HDF5Storage]:
        """If the database is empty, initializes it with the most up to date schema.
        If the database is not empty, ensures that is up to date."""
        if self._db_is_empty():
            Base.metadata.create_all(self._engine)
            self._alembic_util.stamp_head()
        else:
            if not self._alembic_util.db_is_up_to_date():
                path = str(self._engine.url)
                raise EntropyError(
                    f"The database at {path} is not up-to-date. Update the database "
                    "using the Entropy CLI command `entropy upgrade`. "
                    "* Before upgrading be sure to back up your database to a safe place *."
                )
        return self._engine, self._storage

    @staticmethod
    def _print_project_created(path):
        print(
            f"New Entropy project '{project_name(path)}' created at '{project_path(path)}'"
        )

    @staticmethod
    def _validate_path(path):
        if path is None or path == _SQL_ALCHEMY_MEMORY:
            return
        if path is not None and Path(path).suffix == ".db":
            logger.error(
                "_DbInitializer provided with path to a sqlite database. This is deprecated."
            )
            raise EntropyError(
                "Providing the SqlAlchemyDB() constructor with a path to a sqlite "
                "database is deprecated. You should instead provide the path to a "
                "directory containing an Entropy project.\n"
                "To upgrade your existing sqlite database file to an Entropy project "
                "please use the Entropy CLI command: `entropy upgrade`.\n"
                "* Before upgrading be sure to back up your database to a safe place *."
            )
        if path is not None and os.path.isfile(path):
            logger.error(
                "_DbInitializer provided with path to a file, not a directory."
            )
            raise RuntimeError(
                "SqlAlchemyDB() constructor provided with a path to a file but "
                "expects the path to an Entropy project directory"
            )

    def _db_is_empty(self) -> bool:
        cursor = self._engine.execute(
            "SELECT sql FROM sqlite_master WHERE type = 'table'"
        )
        return len(cursor.fetchall()) == 0


class _DbUpgrader:
    def __init__(self, path: str, echo=False) -> None:
        self._path = path
        self._echo = echo
        self._engine = None
        self._storage = None
        self._alembic_util = None

    def upgrade_db(self) -> None:
        if self._path is None or self._path == _SQL_ALCHEMY_MEMORY:
            logger.debug("_DbUpgrader is in in-memory mode")
            self._storage = HDF5Storage()
            self._engine = create_engine(
                "sqlite:///" + _SQL_ALCHEMY_MEMORY, echo=self._echo
            )
        else:
            logger.debug("_DbUpgrader is in project directory mode")
            if not os.path.exists(self._path):
                raise EntropyError(f"No Entropy project exists at '{self._path}'")
            if self._path_is_to_db():
                self._convert_to_project()
            entropy_dir_path = os.path.join(self._path, _ENTROPY_DIRNAME)
            db_file_path = os.path.join(entropy_dir_path, _DB_FILENAME)
            hdf5_file_path = os.path.join(entropy_dir_path, _HDF5_FILENAME)
            self._engine = create_engine("sqlite:///" + db_file_path, echo=self._echo)
            self._storage = HDF5Storage(hdf5_file_path)
        self._alembic_util = _AlembicUtil(self._engine)
        self._alembic_util.upgrade()
        self._migrate_results_to_hdf5()
        self._migrate_metadata_to_hdf5()

    def _path_doesnt_exist(self):
        return not os.path.isdir(self._path) and not os.path.isfile(self._path)

    def _path_is_to_db(self):
        return (
            self._path is not None
            and (Path(self._path).suffix == ".db")
            or os.path.isfile(self._path)
        )

    def _convert_to_project(self):
        # old
        old_db_file_path = self._path
        old_hdf5_file_path = old_db_file_path.replace(".db", ".hdf5")
        # new
        new_project_dir_path = os.path.splitext(old_db_file_path)[0]
        new_entropy_dir_path = os.path.join(new_project_dir_path, _ENTROPY_DIRNAME)
        new_db_file_path = os.path.join(new_entropy_dir_path, _DB_FILENAME)
        new_hdf5_file_path = os.path.join(new_entropy_dir_path, _HDF5_FILENAME)
        # convert
        os.makedirs(new_entropy_dir_path, exist_ok=True)
        shutil.move(old_db_file_path, new_db_file_path)
        if os.path.exists(old_hdf5_file_path):
            shutil.move(old_hdf5_file_path, new_hdf5_file_path)
        self._path = new_project_dir_path
        print(
            f"Converted db file at {old_db_file_path} to project directory "
            f"at {new_project_dir_path}"
        )

    def _migrate_results_to_hdf5(self) -> None:
        self._migrate_rows_to_hdf5(EntityType.RESULT, ResultTable)

    def _migrate_metadata_to_hdf5(self) -> None:
        self._migrate_rows_to_hdf5(EntityType.METADATA, MetadataTable)

    def _migrate_rows_to_hdf5(self, entity_type: EntityType, table: Type[T]):
        logger.debug(f"Migrating {entity_type.name} rows from sqlite to hdf5")
        session_maker = sessionmaker(bind=self._engine)
        with session_maker() as session:
            rows = session.query(table).filter(table.saved_in_hdf5.is_(False)).all()
            if len(rows) == 0:
                logger.debug(f"No {entity_type.name} rows need migrating. Done")
            else:
                logger.debug(f"Found {len(rows)} {entity_type.name} rows to migrate")
                self._storage.migrate_rows(entity_type, rows)
                logger.debug(f"Migrated {len(rows)} {entity_type.name} to hdf5")
                for row in rows:
                    row.saved_in_hdf5 = True
                session.commit()
                logger.debug(
                    f"Marked all {entity_type.name} rows in sqlite as `saved_in_hdf5`. Done"
                )


class _AlembicUtil:
    def __init__(self, engine: sqlalchemy.engine.Engine):
        self._engine = engine

    @staticmethod
    def _abs_path_to(rel_path: str) -> str:
        source_path = Path(__file__).resolve()
        source_dir = source_path.parent
        return os.path.join(source_dir, rel_path)

    def upgrade(self) -> None:
        with self._engine.connect() as connection:
            alembic_cfg = self._alembic_build_config(connection)
            command.upgrade(alembic_cfg, "head")

    def stamp_head(self) -> None:
        with self._engine.connect() as connection:
            alembic_cfg = self._alembic_build_config(connection)
            command.stamp(alembic_cfg, "head")

    def db_is_up_to_date(self) -> bool:
        script_location = self._abs_path_to("alembic")
        script_ = script.ScriptDirectory(script_location)
        with self._engine.begin() as conn:
            context = migration.MigrationContext.configure(conn)
            db_version = context.get_current_revision()
            latest_version = script_.get_current_head()
            return db_version == latest_version

    def _alembic_build_config(self, connection: sqlalchemy.engine.Connection) -> Config:
        config_location = self._abs_path_to("alembic.ini")
        script_location = self._abs_path_to("alembic")
        alembic_cfg = Config(config_location)
        alembic_cfg.set_main_option("script_location", script_location)
        """
        Modified by @urig to share a single engine across multiple (typically in-memory)
        connections, based on this cookbook recipe:
        https://alembic.sqlalchemy.org/en/latest/cookbook.html#connection-sharing
        """
        alembic_cfg.set_main_option("sqlalchemy.url", "sqlite:///:memory:")
        alembic_cfg.attributes["connection"] = connection  # overrides dummy url above
        return alembic_cfg



from datetime import datetime
from time import time_ns
from typing import List, Optional, Iterable, Any, Dict, Tuple

from pandas import DataFrame

from entropylab.api.data_reader import (
    DataReader,
    ResultRecord,
    DebugRecord,
    MetadataRecord,
    ExperimentRecord,
    ScriptViewer,
    PlotRecord,
)
from entropylab.api.data_writer import DataWriter, PlotSpec, NodeData
from entropylab.api.data_writer import (
    ExperimentInitialData,
    ExperimentEndData,
    RawResultData,
    Metadata,
    Debug,
)


class MemoryOnlyDataReaderWriter(DataWriter, DataReader):
    """
    Implementation of DataWriter and DataReader that saves all the data
    to objects in memory.
    Used if no other implementation of the db is used in entropy.
    """

    def __init__(self):
        super(DataWriter, self).__init__()
        super(DataReader, self).__init__()
        self._initial_data: Optional[ExperimentInitialData] = None
        self._end_data: Optional[ExperimentEndData] = None
        self._results: List[Tuple[RawResultData, datetime]] = []
        self._metadata: List[Tuple[Metadata, datetime]] = []
        self._debug: Optional[Debug] = None
        self._plot: Dict[PlotSpec, Any] = {}
        self._nodes: List[NodeData] = []

    def save_experiment_initial_data(self, initial_data: ExperimentInitialData) -> int:
        self._initial_data = initial_data
        return time_ns()

    def save_experiment_end_data(self, experiment_id: int, end_data: ExperimentEndData):
        self._end_data = end_data

    def save_result(self, experiment_id: int, result: RawResultData):
        self._results.append((result, datetime.now()))

    def save_metadata(self, experiment_id: int, metadata: Metadata):
        self._metadata.append((metadata, datetime.now()))

    def save_debug(self, experiment_id: int, debug: Debug):
        self._debug = debug

    def save_plot(self, experiment_id: int, plot: PlotSpec, data: Any):
        self._plot[plot] = data

    def save_node(self, experiment_id: int, node_data: NodeData):
        self._nodes.append(node_data)

    def get_experiments_range(self, starting_from_index: int, count: int) -> DataFrame:
        raise NotImplementedError()

    def get_experiments(
        self,
        label: Optional[str] = None,
        start_after: Optional[datetime] = None,
        end_after: Optional[datetime] = None,
        success: Optional[bool] = None,
    ) -> Iterable[ExperimentRecord]:
        raise NotImplementedError()

    def get_experiment_record(self, experiment_id: int) -> Optional[ExperimentRecord]:
        if self._initial_data:
            if self._end_data:
                end_time = self._end_data.end_time
            else:
                end_time = None
            return ExperimentRecord(
                experiment_id,
                self._initial_data.label,
                ScriptViewer([self._initial_data.script]),
                self._initial_data.start_time,
                end_time,
                self._initial_data.story,
                self._end_data.success,
            )
        else:
            return None

    def get_results(
        self,
        experiment_id: Optional[int] = None,
        label: Optional[str] = None,
        stage: Optional[int] = None,
    ) -> Iterable[ResultRecord]:
        return list(
            ResultRecord(
                experiment_id,
                self._results.index(x),
                x[0].label,
                x[0].story,
                x[0].stage,
                x[0].data,
                x[1],
            )
            for x in self._results
            if (not label or x[0].label == label)
            and (stage is None or x[0].stage == stage)
        )

    def get_metadata_records(
        self,
        experiment_id: Optional[int] = None,
        label: Optional[str] = None,
        stage: Optional[int] = None,
    ) -> Iterable[MetadataRecord]:
        return list(
            MetadataRecord(
                experiment_id,
                self._metadata.index(x),
                x[0].label,
                x[0].stage,
                x[0].data,
                x[1],
            )
            for x in self._metadata
            if (not label or x[0].label == label)
            and (stage is None or x[0].stage == stage)
        )

    def get_debug_record(self, experiment_id: int) -> Optional[DebugRecord]:
        if self._debug:
            return DebugRecord(
                experiment_id,
                0,
                self._debug.python_env,
                self._debug.python_history,
                self._debug.station_specs,
                self._debug.extra,
            )
        else:
            return None

    def get_plots(self, experiment_id: int) -> List[PlotRecord]:
        return [
            PlotRecord(
                experiment_id,
                id(plot),
                self._plot[plot],
                plot.generator(),
                plot.label,
                plot.story,
            )
            for plot in self._plot
        ]

    def get_nodes_id_by_label(
        self, label: str, experiment_id: Optional[int] = None
    ) -> List[int]:
        return list(x.node_id for x in self._nodes if (not label or x.label == label))

    def get_last_result_of_experiment(
        self, experiment_id: int
    ) -> Optional[ResultRecord]:
        if len(self._results) > 0:
            raw_result = self._results[-1]
            if raw_result:
                index = self._results.index(raw_result)
                return ResultRecord(
                    experiment_id,
                    index,
                    raw_result[0].label,
                    raw_result[0].story,
                    raw_result[0].stage,
                    raw_result[0].data,
                    raw_result[1],
                )
            else:
                return None
        else:
            return None



import abc
from typing import List, Dict

# import numpy as np
import pandas as pd
from pandas import DataFrame

from entropylab import SqlAlchemyDB
from entropylab.api.data_reader import PlotRecord
from entropylab.results.dashboard.auto_plot import auto_plot


class DashboardDataReader(abc.ABC):
    @abc.abstractmethod
    def get_last_experiments(self, number) -> DataFrame:
        pass

    @abc.abstractmethod
    def get_plot_data(self, exp_id: int) -> List[PlotRecord]:
        pass


class SqlalchemyDashboardDataReader(DashboardDataReader):
    def __init__(self, connector: SqlAlchemyDB) -> None:
        super().__init__()
        self._db: SqlAlchemyDB = connector

    def get_last_experiments(self, max_num_of_experiments: int) -> List[Dict]:
        experiments = self._db.get_last_experiments(max_num_of_experiments)
        experiments["success"] = experiments["success"].apply(
            lambda x: "✔️" if x else "❌"
        )
        experiments["start_time"] = pd.DatetimeIndex(
            experiments["start_time"]
        ).strftime("%Y-%m-%d %H:%M:%S")
        experiments["end_time"] = pd.DatetimeIndex(experiments["end_time"]).strftime(
            "%Y-%m-%d %H:%M:%S"
        )
        records = experiments.to_dict("records")
        return records

    def get_last_result_of_experiment(self, experiment_id):
        return self._db.get_last_result_of_experiment(experiment_id)

    def get_plot_data(self, exp_id: int) -> List[PlotRecord]:
        plots = self._db.get_plots(exp_id)
        if len(plots) > 0:
            return plots
        else:
            last_result = self._db.get_last_result_of_experiment(exp_id)
            if last_result is not None and last_result.data is not None:
                plot = auto_plot(exp_id, last_result.data)
                return [plot]
            else:
                return []



from __future__ import annotations

import hashlib
import json
import threading
import time
from datetime import datetime
from typing import Optional, Dict, Any, List, Callable

import pandas as pd
from munch import Munch
from tinydb import TinyDB, Query
from tinydb.storages import MemoryStorage
from tinydb.table import Document

from entropylab.api.errors import EntropyError
from entropylab.api.param_store import ParamStore, MergeStrategy


class InProcessParamStore(ParamStore, Munch):

    """Naive implementation of ParamStore based on tinydb

    Important:
    Using this implementation in multiple concurrent processes is not supported.
    """

    def __init__(
        self,
        path: Optional[str] = None,
        theirs: Optional[Dict | ParamStore] = None,
        merge_strategy: Optional[MergeStrategy] = MergeStrategy.THEIRS,
    ):

        super().__init__()

        self.__lock = threading.RLock()
        self.__base_commit_id: Optional[str] = None  # last commit checked out/committed
        self.__base_doc_id: Optional[int] = None  # tinydb document id of last commit...
        self.__tags: Dict[str, List[str]] = dict()  # tags that are mapped to keys
        self.__is_dirty: bool = True  # can the store be committed at this time?

        if path is None:
            self.__is_in_memory_mode = True
            self.__db = TinyDB(storage=MemoryStorage)
        else:
            self.__is_in_memory_mode = False
            self.__db = TinyDB(path)
        if theirs is not None:
            self.merge(theirs, merge_strategy)

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.__db.close()

    """ Properties """

    @property
    def is_dirty(self):
        with self.__lock:
            return self.__is_dirty

    """ MutableMapping """

    def __setitem__(self, key: str, value: Any) -> None:
        """
        Set self[key] to value. The key-value pair becomes a "param" and
        can be persisted using `commit()` and retrieved later using
        `checkout()`.

        Note: Keys should not start with a dunder (`__`). Such keys are not
        treated as params and are not persisted when `commit()` is called.
        """

        if key.startswith("__") or key.startswith(f"_{self.__class__.__name__}__"):
            # keys that are private attributes are not params and are treated
            # as regular object attributes
            object.__setattr__(self, key, value)
        else:
            with self.__lock:
                super().__setitem__(key, value)
                self.__is_dirty = True

    def __getitem__(self, key: str) -> Any:
        with self.__lock:
            return super().__getitem__(key)

    def __delitem__(self, *args, **kwargs):
        with self.__lock:
            super().__delitem__(*args, **kwargs)
            self.__remove_key_from_tags(args[0])
            self.__is_dirty = True

    def to_dict(self) -> Dict:
        with self.__lock:
            return super().toDict()

    def get(self, key: str, commit_id: Optional[str] = None):
        with self.__lock:
            if commit_id is None:
                super().__getitem__(key)
            else:
                commit = self.__get_commit(commit_id)
                return commit["params"][key]

    def __remove_key_from_tags(self, key: str):
        for tag in self.__tags:
            if key in self.__tags[tag]:
                self.__tags[tag].remove(key)

    """ Commits """

    def commit(self, label: Optional[str] = None) -> str:
        with self.__lock:
            if not self.__is_dirty:
                return self.__base_commit_id
            doc = self.__build_document(label)
            doc_id = self.__db.insert(doc)
            self.__base_commit_id = doc["metadata"]["id"]
            self.__base_doc_id = doc_id
            self.__is_dirty = False
            return doc["metadata"]["id"]

    def __build_document(self, label: Optional[str] = None) -> dict:
        metadata = self.__build_metadata(label)
        params = self.toDict()
        return Munch(metadata=metadata.__dict__, params=params, tags=self.__tags)

    def __build_metadata(self, label: Optional[str] = None) -> Metadata:
        metadata = Metadata()
        metadata.ns = time.time_ns()
        params_json = json.dumps(self.toDict(), sort_keys=True, ensure_ascii=True)
        commit_encoded = (params_json + str(metadata.ns)).encode("utf-8")
        metadata.id = hashlib.sha1(commit_encoded).hexdigest()
        metadata.label = label
        return metadata

    def checkout(
        self,
        commit_id: Optional[str] = None,
        commit_num: Optional[int] = None,
        move_by: Optional[int] = None,
    ) -> None:
        with self.__lock:
            commit = self.__get_commit(commit_id, commit_num, move_by)
            self.clear()
            self.update(commit["params"])
            self.__tags = commit["tags"]
            self.__base_commit_id = commit_id
            self.__base_doc_id = commit.doc_id
            self.__is_dirty = False

    def list_commits(self, label: Optional[str] = None) -> List[Metadata]:
        with self.__lock:
            documents = self.__db.search(
                Query().metadata.label.test(_test_if_value_contains(label))
            )
            metadata = map(_extract_metadata, documents)
            return list(metadata)

    def __get_commit(
        self,
        commit_id: Optional[str] = None,
        commit_num: Optional[int] = None,
        move_by: Optional[int] = None,
    ) -> Document:
        if commit_id is not None:
            commit = self.__get_commit_by_id(commit_id)
        elif commit_num is not None:
            commit = self.__get_commit_by_num(commit_num)
        elif move_by is not None:
            commit = self.__get_commit_by_move_by(move_by)
        else:
            raise EntropyError(
                "Please provide one of the following arguments: "
                "commit_id, commit_num, or move_by"
            )
        return commit

    def __get_commit_by_id(self, commit_id: str):
        result = self.__db.search(Query().metadata.id == commit_id)
        if len(result) == 0:
            raise EntropyError(f"Commit with id '{commit_id}' not found")
        if len(result) > 1:
            raise EntropyError(
                f"{len(result)} commits with id '{commit_id}' found. "
                f"Only one commit is allowed per id"
            )
        return result[0]

    def __get_commit_by_num(self, commit_num: int):
        result = self.__db.get(doc_id=commit_num)
        if result is None:
            raise EntropyError(f"Commit with number '{commit_num}' not found")
        return result

    def __get_commit_by_move_by(self, move_by: int):
        doc_id = self.__base_doc_id + move_by
        result = self.__db.get(doc_id=doc_id)
        return result

    """ Merge """

    def merge(
        self,
        theirs: Dict | ParamStore,
        merge_strategy: Optional[MergeStrategy] = MergeStrategy.OURS,
    ) -> None:
        if issubclass(type(theirs), ParamStore):
            theirs = theirs.to_dict()
        with self.__lock:
            ours = self.toDict()
            merged = self.__merge_trees(ours, theirs, merge_strategy)
            self.clear()
            self.update(merged)

    def __merge_trees(self, a: Dict, b: Dict, merge_strategy: MergeStrategy) -> Dict:
        """Merges dict b into dict a using the given strategy"""
        for key in b:
            if key in a:
                if isinstance(a[key], dict) and isinstance(b[key], dict):
                    self.__merge_trees(a[key], b[key], merge_strategy)
                elif a[key] == b[key]:
                    pass  # same leaf value, nothing to do
                else:  # conflict:
                    if merge_strategy == MergeStrategy.OURS:
                        pass  # a takes precedence, ignore b
                    elif merge_strategy == MergeStrategy.THEIRS:
                        a[key] = b[key]  # b takes precedence, overwrite a
                        self.__is_dirty = True
                    else:
                        raise NotImplementedError(
                            f"MergeStrategy '{merge_strategy}' is not implemented"
                        )
            else:  # key from b is not in a:
                a[key] = b[key]  # "copy" from b to a
                self.__is_dirty = True
        return a

    def list_values(self, key: str) -> pd.DataFrame:
        with self.__lock:
            values = []
            commits = self.__db.all()
            commits.sort(key=lambda x: x["metadata"]["ns"])
            for commit in commits:
                try:
                    value = (
                        commit["params"][key],
                        _ns_to_datetime(commit["metadata"]["ns"]),
                        commit["metadata"]["id"],
                        commit["metadata"]["label"],
                    )
                    values.append(value)
                except KeyError:
                    pass
            if key in self.keys():
                values.append((self.__getitem__(key), None, None, None))
            df = pd.DataFrame(values)
            if not df.empty:
                df.columns = ["value", "time", "commit_id", "label"]
            return df

    """ Tags """

    def add_tag(self, tag: str, key: str) -> None:
        with self.__lock:
            if key not in self.keys():
                raise KeyError(f"key '{key}' is not in store")
            if tag not in self.__tags:
                self.__tags[tag] = []
            self.__tags[tag].append(key)
            self.__is_dirty = True

    def remove_tag(self, tag: str, key: str) -> None:
        with self.__lock:
            if tag not in self.__tags:
                return
            if key not in self.__tags[tag]:
                return
            self.__tags[tag].remove(key)
            self.__is_dirty = True

    def list_keys(self, tag: str) -> List[str]:
        with self.__lock:
            if tag not in self.__tags:
                return []
            else:
                return self.__tags[tag]

    """ Temporary State """

    def save_temp(self) -> None:
        """
        Saves the state of params to a temporary location
        """
        with self.__lock:
            table = self.__db.table("temp")
            doc = self.__build_document()
            table.upsert(Document(doc, doc_id=1))

    def load_temp(self) -> None:
        """
        Overwrites the current state of params with data loaded from the temporary
        location
        """
        with self.__lock:
            table = self.__db.table("temp")
            doc = table.get(doc_id=1)
            if not doc:
                raise EntropyError(
                    "Temp location is empty. Call save_temp() before calling load_temp()"
                )
            self.clear()
            self.update(doc["params"])
            self.__tags = doc["tags"]


class Metadata:
    id: str
    ns: int
    label: Optional[str]

    def __init__(self, d: Dict = None):
        if d:
            self.__dict__.update(d)

    def __repr__(self) -> str:
        d = self.__dict__.copy()
        d["ns"] = _ns_to_datetime(self.ns)
        return f"<Metadata({_dict_to_json(d)})>"


""" Static helper methods """


def _test_if_value_contains(label: str) -> Callable:
    return lambda val: (label or "") in (val or "")


def _extract_metadata(document: Document) -> Metadata:
    return Metadata(document.get("metadata"))


def _dict_to_json(d: Dict) -> str:
    return json.dumps(d, default=_json_dumps_default, sort_keys=True, ensure_ascii=True)


def _json_dumps_default(value):
    if isinstance(value, datetime):
        return str(value)
    else:
        return value.__dict__


def _ns_to_datetime(ns: int) -> pd.datetime:
    """Convert a UNIX epoch timestamp in nano-seconds to a human readable string"""
    return pd.to_datetime(ns)



import abc
import enum
import importlib
import inspect
import pickle
from dataclasses import dataclass
from typing import Type, Optional, Dict, Any, Iterable, List, Set

from entropylab.api.data_reader import DataReader
from entropylab.api.data_writer import DataWriter
from entropylab.api.errors import ResourceNotFound, EntropyError
from entropylab.instruments.instrument_driver import (
    Resource,
    Function,
    Parameter,
)
from entropylab.logger import logger


class DriverType(enum.Enum):
    Packaged = 1
    Local = 2


@dataclass
class ResourceRecord:
    name: str
    module: str
    class_name: str
    driver_code: str
    version: str
    driver_type: DriverType
    args: str
    kwargs: str
    cached_functions: List[Function]
    cached_parameters: List[Parameter]
    cached_undeclared_functions: List[Function]


class PersistentLabDB(abc.ABC):
    """
    An abstract class for Entropy database, defines the way entropy saved lab data
    A subclass of this class will enable saving and sharing the lab between
    different executions and experiments
    """

    def __init__(self) -> None:
        super().__init__()

    @abc.abstractmethod
    def save_new_resource_driver(
        self,
        name: str,
        driver_source_code: str,
        module: str,
        class_name: str,
        serialized_args: str,
        serialized_kwargs: str,
        number_of_experiment_args: int,
        keys_of_experiment_kwargs: List[str],
        functions: List[Function],
        parameters: List[Parameter],
        undeclared_functions: List[Function],
    ):
        """
            saves all required data for restoring a resource
        :param name: resource name, should be unique
        :param driver_source_code: driver source code for debugging
        :param module: driver python module
        :param class_name: driver python class
        :param serialized_args: arguments for initializing the driver
        :param serialized_kwargs: key word arguments for initializing the driver
        :param number_of_experiment_args: the number of arguments that are passed
                                            on every request
        :param keys_of_experiment_kwargs: the keys of key word arguments that are
                                        passed on every request
        :param functions: cached list of functions
        :param undeclared_functions: cached list of undeclared functions
        :param parameters:  cached list of parameters
        """
        pass

    @abc.abstractmethod
    def remove_resource(
        self,
        name: str,
    ):
        """
            remove a resource from the persistent lab
        :param name: resource name
        """
        pass

    @abc.abstractmethod
    def save_state(self, name: str, state: str, snapshot_name: str):
        """
            save a specific state/snapshot of the given resource
            If resource implemented the revert_to_snapshot function,
            it can be used for restoring this state to the resource.
        :param name: resource name
        :param state: serialized state
        :param snapshot_name: the name of the snapshot for future use.
                            should be unique for a given resource.
        """
        pass

    @abc.abstractmethod
    def get_state(self, resource_name: str, snapshot_name: str) -> str:
        """
            return resource serialized state
        :param resource_name: the name of the resource
        :param snapshot_name: the name this state is tagged.
        """
        pass

    @abc.abstractmethod
    def get_all_states(self, name) -> Iterable[str]:
        """
            returns a list of all resource serialized states
        :param name:
        """
        pass

    @abc.abstractmethod
    def get_resource(self, name) -> Optional[ResourceRecord]:
        """
            return information about a specific resource
        :param name: resource name
        """
        pass

    @abc.abstractmethod
    def set_locked(self, resource_name):
        """
            sets the given resource as locked, meaning no one else can use this resource
            until it released.
        :param resource_name: the name of the resource
        """
        pass

    @abc.abstractmethod
    def set_released(self, resource_name):
        """
            release the given resource, meaning now some one else can lock the resource
            and use it.
        :param resource_name:
        """
        pass

    @abc.abstractmethod
    def get_all_resources(self) -> Set[str]:
        """
        :returns a list of all saved resources names
        """
        pass


def _get_class(module_name, class_name):
    module = importlib.import_module(module_name)
    if not hasattr(module, class_name):
        raise Exception("class {} is not in {}".format(class_name, module_name))
    logger.debug("reading class {} from module {}".format(class_name, module_name))
    cls = getattr(module, class_name)
    return cls


@dataclass
class _LabTopologyResourceInstance:
    name: str
    resource_class: Optional[Type]
    args: Optional[Any]
    kwargs: Optional[Any]
    resource: Optional[Any]


class LabResources:
    """
    A class for managing a persistent lab.
    """

    def __init__(self, persistence_db: PersistentLabDB) -> None:
        """
            A class for managing a persistent lab.
        :param persistence_db: a database that implemented PersistentLabDB interface.
                                the db can be the same as results db, or
                                a different one. This db can be shared across
                                executions and users,
        """
        super().__init__()
        self._resources: Dict[str, _LabTopologyResourceInstance] = {}
        self._persistent_db = persistence_db

    def resource_exist(self, name: str) -> bool:
        """
            True if resource exists in this lab
        :param name: resource unique name
        :return:
        """
        if name in self._resources:
            return True
        else:
            return self._persistent_db.get_resource(name) is not None

    def get_resource(
        self,
        name: str,
        experiment_args: Optional[List] = None,
        experiment_kwargs: Optional[Dict] = None,
    ):
        """
            returns an initialized instance of the resource,
            ready to execute actions.
        :param name: resource name
        :param experiment_args: arguments for initializing the resource,
                                as specified in register.
        :param experiment_kwargs: key word arguments for initializing the
                                resource, as specified in register.
        """
        if name not in self._resources:
            loaded_resource = self._get_instance(
                name,
                experiment_args=experiment_args,
                experiment_kwargs=experiment_kwargs,
            )
            self._resources[name] = loaded_resource
            return loaded_resource.resource
        else:
            return self._resources[name].resource

    def _get_resource_if_already_initialized(self, name: str):
        if name in self._resources:
            return self._resources[name].resource
        else:
            raise ResourceNotFound()

    def _get_instance(
        self,
        name,
        experiment_args: Optional[List] = None,
        experiment_kwargs: Optional[Dict] = None,
    ):
        if name in self._resources:
            logger.debug(f"got device {name} from memory")
            return self._resources[name]
        else:
            record = self._persistent_db.get_resource(name)
            if record:
                args = pickle.loads(record.args)
                kwargs = pickle.loads(record.kwargs)
                resource_class = _get_class(record.module, record.class_name)
                if args is None:
                    args = []
                if kwargs is None:
                    kwargs = {}
                if experiment_args is None:
                    experiment_args = []
                if experiment_kwargs is None:
                    experiment_kwargs = {}
                combined_args = list(args) + list(experiment_args)
                combined_kwargs = {**kwargs, **experiment_kwargs}
                instance = resource_class(*combined_args, **combined_kwargs)
                if isinstance(instance, Resource):
                    instance.set_entropy_name(name)
                return _LabTopologyResourceInstance(
                    record.name,
                    resource_class,
                    combined_args,
                    combined_kwargs,
                    instance,
                )

        raise ResourceNotFound()

    def register_resource(
        self,
        name,
        resource_class: Type,
        args: Optional[List] = None,
        experiment_args: Optional[List] = None,
        kwargs: Optional[Dict] = None,
        experiment_kwargs: Optional[Dict] = None,
        dynamic_driver_specs_discovery: bool = False,
    ):
        """
            register a new resource and it's driver to the persistent lab.
            any combination of args and kwargs can be given.
        :param name: resource unique name
        :param resource_class: driver python class
        :param args: arguments for initializing the driver,
                    will be serialized and saved to the persistent lab
        :param experiment_args: more arguments for initializing the driver,
                                won't be save to the persistent lab, and
                                should be specified on every import.
        :param kwargs: key word arguments for initializing the driver,
                    will be serialized and saved to the persistent lab
        :param experiment_kwargs: more keyword arguments for initializing
                                the driver, won't be save to the persistent lab, and
                                should be specified on every import.
        :param dynamic_driver_specs_discovery: whether to run active discovery of
                                                driver functionality
        """
        if name in self._resources or self.resource_exist(name):
            raise KeyError(f"instrument {name} already exist")
        self.update_resource(
            name,
            resource_class,
            args,
            experiment_args,
            kwargs,
            experiment_kwargs,
            dynamic_driver_specs_discovery,
        )

    def register_resource_if_not_exist(
        self,
        name,
        resource_class: Type,
        args: Optional[List] = None,
        experiment_args: Optional[List] = None,
        kwargs: Optional[Dict] = None,
        experiment_kwargs: Optional[Dict] = None,
    ):
        """
            register a new resource and it's driver to the persistent lab,
            only if it's not already exist.
            any combination of args and kwargs can be given.
        :param name: resource unique name
        :param resource_class: driver python class
        :param args: arguments for initializing the driver,
                    will be serialized and saved to the persistent lab
        :param experiment_args: more arguments for initializing the driver,
                                won't be save to the persistent lab, and
                                should be specified on every import.
        :param kwargs: key word arguments for initializing the driver,
                    will be serialized and saved to the persistent lab
        :param experiment_kwargs: more keyword arguments for initializing
                                the driver, won't be save to the persistent lab, and
                                should be specified on every import.
        """
        if name not in self._resources and not self.resource_exist(name):
            self.register_resource(
                name,
                resource_class,
                args,
                experiment_args=experiment_args,
                kwargs=kwargs,
                experiment_kwargs=experiment_kwargs,
            )

    def save_snapshot(self, resource_name: str, snapshot_name: str):
        """
        save the current state of the given resource to persistent lab.
        the state will be tagged with the snapshot name,
        and can be requested later.
        """
        if resource_name not in self._resources:
            raise ResourceNotFound()

        device = self._resources[resource_name]
        is_entropy_resource = issubclass(device.resource_class, Resource)
        if is_entropy_resource:
            str_snapshot = device.resource.snapshot(False)
        else:
            raise TypeError(
                "resource is not an entropy resource and snapshot can't be saved"
            )
        self._persistent_db.save_state(device.name, str_snapshot, snapshot_name)

    def all_resources(self) -> Set[str]:
        """
            returns a set of all resource names in the lab
        :return:
        """
        return self._persistent_db.get_all_resources()

    def lock_resources(self, resources_names: Set[str]):
        """
            Lock all the given resources, so no one else will use them
        :param resources_names: a set of all resources to be locked
        """
        for resource_name in resources_names:
            resource_record = self._persistent_db.get_resource(resource_name)
            if resource_record:
                self._persistent_db.set_locked(resource_name)
            else:
                raise ResourceNotFound()

    def release_resources(self, resources_names: Set[str]):
        """
            Release the given resources for others to use
        :param resources_names: a set of all resources to be released
        """
        failed = False
        for resource_name in resources_names:
            try:
                if resource_name in self._resources:
                    resource = self._resources[resource_name]
                    if isinstance(resource, Resource):
                        resource.teardown()
                    self._persistent_db.set_released(resource_name)
            except Exception:
                failed = True

        if failed:
            raise Exception("failed releasing some resource")

    def get_resource_info(self, resource_name) -> Dict:
        """
        returns information about the given resource
        """
        resource_record = self._persistent_db.get_resource(resource_name)
        if resource_record:
            return {"resource_class": resource_record.class_name}
        else:
            raise ResourceNotFound()

    def get_snapshot(self, resource_name: str, snapshot_name: str) -> str:
        """
            returns a serialized state of the given resource.
        :param resource_name: the name of the resource
        :param snapshot_name: the tag of the requested state
        :return:
        """
        return self._persistent_db.get_state(resource_name, snapshot_name)

    def remove_resource(self, resource_name):
        """
        remove the given resource from the persistent lab.
        """
        self._persistent_db.remove_resource(resource_name)

    def update_resource(
        self,
        name,
        resource_class: Type,
        args: Optional[List] = None,
        experiment_args: Optional[List] = None,
        kwargs: Optional[Dict] = None,
        experiment_kwargs: Optional[Dict] = None,
        dynamic_driver_specs_discovery: bool = False,
    ):
        """
            update the given resource with a new driver
        :param name: resource unique name
        :param resource_class: driver python class
        :param args: arguments for initializing the driver,
                    will be serialized and saved to the persistent lab
        :param experiment_args: more arguments for initializing the driver,
                                won't be save to the persistent lab, and
                                should be specified on every import.
        :param kwargs: key word arguments for initializing the driver,
                    will be serialized and saved to the persistent lab
        :param experiment_kwargs: more keyword arguments for initializing
                                the driver, won't be save to the persistent lab, and
                                should be specified on every import.
        :param dynamic_driver_specs_discovery: whether to run active discovery of
                                                driver functionality
        """
        is_entropy_resource = issubclass(resource_class, Resource)
        if not is_entropy_resource:
            logger.warn(
                f"instrument {name} is not an entropylab Resource and"
                f" additional metadata won't be saved"
            )

        if args is None:
            args = []
        if kwargs is None:
            kwargs = {}
        if experiment_args is None:
            experiment_args = []
        if experiment_kwargs is None:
            experiment_kwargs = {}
        combined_args = list(args) + list(experiment_args)
        combined_kwargs = {**kwargs, **experiment_kwargs}
        number_of_experiment_args = len(experiment_args)
        keys_of_experiment_kwargs = kwargs.keys()

        logger.debug(f"Initialize device {name}")
        instance = resource_class(*combined_args, **combined_kwargs)
        serialized_args = pickle.dumps(args)
        serialized_kwargs = pickle.dumps(kwargs)
        if isinstance(instance, Resource) and dynamic_driver_specs_discovery:
            driver_spec = instance.get_dynamic_driver_specs()
            functions = driver_spec.functions
            undeclared_functions = driver_spec.undeclared_functions
            parameters = driver_spec.parameters
        else:
            functions = []
            parameters = []
            undeclared_functions = []
        del instance
        try:
            source = inspect.getsource(inspect.getmodule(resource_class))
        except Exception:
            logger.warn(f"could not save source of {resource_class.__qualname__}")
            source = ""
        self._persistent_db.save_new_resource_driver(
            name,
            source,
            resource_class.__module__,
            resource_class.__qualname__,
            serialized_args,
            serialized_kwargs,
            number_of_experiment_args,
            list(keys_of_experiment_kwargs),
            functions,
            parameters,
            undeclared_functions,
        )


@dataclass()
class _ExperimentResource:
    resource: Any
    snapshot_name: Optional[str] = None


class ExperimentResources:
    """
    A class for defining all resources needed for a specific experiment
    """

    def __init__(self, persistent_db: Optional[PersistentLabDB] = None) -> None:
        """
            A class for defining all resources needed for a specific experiment
        :param persistent_db: a persistent lab that this experiment runs in
                                if the db implmentation is also a DataWriter subclass,
                                this db will be used for results as well.
        """
        super().__init__()
        self._save_to_db = True
        self._results_db = None
        if isinstance(persistent_db, DataWriter):
            self._results_db = persistent_db
        if isinstance(persistent_db, PersistentLabDB):
            self._persistent_lab_connector = LabResources(persistent_db)
        self._private_results_db = None
        self._resources: Dict[str, _ExperimentResource] = {}
        self._local_resources: Dict[str, Any] = {}
        self._started = False

    def _get_persistent_lab_connector(self) -> LabResources:
        if not self._persistent_lab_connector:
            raise Exception("persistent lab is not configured for this experiment")
        return self._persistent_lab_connector

    def register_private_results_db(self, db: DataWriter):
        """
            set a different results db for this experiment.
        :param db: the results db
        """
        self._private_results_db = db

    def pause_save_to_results_db(self):
        """
        don't save results in this experiment
        """
        logger.warn(
            "DB saving paused, results will be permanently lost on session close"
        )
        self._save_to_db = False

    def resume_save_to_results_db(self):
        """
        save results in this experiment
        """
        logger.warn("DB saving resumed, results will be safely saved to db")
        self._save_to_db = True

    def get_results_db(self) -> Optional[DataWriter]:
        """
        get the results db used in this experiment
        """
        if not self._save_to_db:
            return None
        if self._private_results_db:
            return self._private_results_db
        return self._results_db

    def get_results_reader(self) -> DataReader:
        """
        :returns a data reader from the current results db
        """
        db = self.get_results_db()
        if isinstance(db, DataReader):
            return db
        else:
            raise EntropyError(
                "the experiment db has not implemented DataReader abstract class"
            )

    def import_lab_resource(
        self,
        name: str,
        experiment_args: Optional[List] = None,
        experiment_kwargs: Optional[Dict] = None,
        snapshot_name: Optional[str] = None,
    ):
        """
            import a resource from the persistent lab to the current experiment
        :param name: resource name
        :param experiment_args: arguments for initializing the resource,
                                as specified in the persistent lab register.
        :param experiment_kwargs: key word arguments for initializing the
                                resource, as specified in the persistent lab register.
        :param snapshot_name: optional to import the resource in a specific state,
                              that was tagged with the snapshot name.
                              will only work if the resource implemented the
                              revert_to_snapshot function
        """
        if name in self._resources:
            raise Exception("can't import resource twice")
        if name in self._local_resources:
            raise Exception(
                "resource with the same name " "was already added as temp resource"
            )
        if self._get_persistent_lab_connector().resource_exist(name):
            resource = self._get_persistent_lab_connector().get_resource(
                name,
                experiment_args=experiment_args,
                experiment_kwargs=experiment_kwargs,
            )
            self._resources[name] = _ExperimentResource(resource, snapshot_name)
            if snapshot_name and not isinstance(resource, Resource):
                raise Exception(
                    "Resource is not an entropy resource and can't be reverted"
                )
        else:
            raise Exception("Resource is not a part of the lab")

    def add_temp_resource(self, name: str, instance):
        """
            add a resource to this experiment, that is not part of the persistent lab.
        :param name: resource name
        :param instance: the python object
        """
        if name in self._resources:
            raise Exception(f"resource {name} is already imported from lab")
        if name not in self._local_resources:
            if isinstance(instance, Resource):
                instance.set_entropy_name(name)
            self._local_resources[name] = instance
        else:
            raise Exception(f"resource {name} already exist")

    def get_resource(self, name):
        """
            get the resource instance that was added/imported for this experiment
        :param name: resource name
        :return: resource instance
        """
        if name in self._resources:
            lab_connector = self._get_persistent_lab_connector()
            resource = lab_connector._get_resource_if_already_initialized(name)
        elif name in self._local_resources:
            resource = self._local_resources[name]
        else:
            raise KeyError(
                "cannot use a resource that wasn't added to experiment resources"
            )

        if isinstance(resource, Resource):
            resource = resource.get_instance()
        return resource

    def has_resource(self, name) -> bool:
        """
            True if the resource wass added or imported for this experiment
        :param name:
        :return:
        """
        return name in self._resources or name in self._local_resources

    def start_experiment(self):
        if self._started:
            raise EntropyError("can not start experiment twice")
        self._started = True

        if self._resources:
            self._get_persistent_lab_connector().lock_resources(
                set([resource_name for resource_name in self._resources])
            )

        # connect to all resources
        for resource_name in self._resources:
            resource = self._resources[resource_name]
            if isinstance(resource, Resource):
                resource.connect()
            snapshot_name = resource.snapshot_name
            if snapshot_name is not None:
                snap = self._get_persistent_lab_connector().get_snapshot(
                    resource_name, snapshot_name
                )
                resource_name = self.get_resource(resource_name)
                resource_name.revert_to_snapshot(snap)

    def end_experiment(self):
        if self._started:
            if self._resources:
                self._get_persistent_lab_connector().release_resources(
                    set([resource_name for resource_name in self._resources])
                )
            if self._local_resources:
                for resource_name in self._local_resources:
                    resource = self._local_resources[resource_name]
                    if isinstance(resource, Resource):
                        resource.teardown()
                    del resource
            self._started = False
        else:
            raise EntropyError("can not end not-started or finished experiment")

    def _serialize_resources_snapshot(self) -> Dict[str, str]:
        snapshots = {}
        for resource_name in self._resources:
            lab = self._get_persistent_lab_connector()
            resource = lab._get_resource_if_already_initialized(resource_name)
            if resource and isinstance(resource, Resource):
                snapshots[resource_name] = resource.snapshot(False)
        return snapshots

    def save_snapshot(self, resource_name: str, snapshot_name: str):
        """
        save the resource current state and tag it with the snapshot name
        """
        if resource_name not in self._resources:
            raise ResourceNotFound()
        self._get_persistent_lab_connector().save_snapshot(resource_name, snapshot_name)



import json
from typing import Dict, List

import dash
import dash_bootstrap_components as dbc
from dash import html, dcc
from dash.dependencies import Input, Output, State, ALL
from plotly import graph_objects as go
from plotly.subplots import make_subplots

from entropylab import SqlAlchemyDB
from entropylab.api.data_reader import PlotRecord
from entropylab.api.errors import EntropyError
from entropylab.logger import logger
from entropylab.results.dashboard.dashboard_data import SqlalchemyDashboardDataReader
from entropylab.results.dashboard.layout import layout
from entropylab.results.dashboard.theme import (
    colors,
    dark_plot_layout,
    theme_stylesheet,
)
from entropylab.results_backend.sqlalchemy.project import project_name, project_path

MAX_EXPERIMENTS_NUM = 10000
EXPERIMENTS_PAGE_SIZE = 6


def build_dashboard_app(proj_path):
    """Initialize the results dashboard Dash app to display an Entropy project

    :param proj_path path where the Entropy project to be used resides."""

    """ Creating and setting up our Dash app """

    _dashboard_data_reader = SqlalchemyDashboardDataReader(SqlAlchemyDB(proj_path))

    def _build_layout():
        records = _dashboard_data_reader.get_last_experiments(MAX_EXPERIMENTS_NUM)
        return layout(proj_path, records)

    _app = dash.Dash(
        __name__,
        external_stylesheets=[theme_stylesheet],
        update_title=None,
        suppress_callback_exceptions=True,
    )
    _app.title = f"Entropy - {project_name(proj_path)} [{project_path(proj_path)}]"
    _app.layout = _build_layout  # See: https://dash.plotly.com/live-updates

    """ CALLBACKS and their helper functions """

    @_app.callback(
        Output("experiments-table", "data"),
        Output("empty-project-modal", "is_open"),
        Input("interval", "n_intervals"),
    )
    def refresh_experiments_table(_):
        """{Periodically refresh the experiments table.
        See https://dash.plotly.com/live-updates}"""
        records = _dashboard_data_reader.get_last_experiments(MAX_EXPERIMENTS_NUM)
        open_empty_project_modal = len(records) == 0
        return records, open_empty_project_modal

    @_app.callback(
        Output("failed-plotting-alert", "is_open"),
        Input("failed-plotting-alert", "children"),
    )
    def open_failed_plotting_alert_when_its_not_empty(children):
        return children != ""

    @_app.callback(
        Output("plot-tabs", "children"),
        Output("plot-figures", "data"),
        Output("prev-selected-rows", "data"),
        Output("failed-plotting-alert", "children"),
        Output("no-paging-spacer", "hidden"),
        Input("experiments-table", "selected_rows"),
        State("experiments-table", "data"),
        State("plot-figures", "data"),
        State("prev-selected-rows", "data"),
    )
    def render_plot_tabs_from_selected_experiments_table_rows(
        selected_rows, data, plot_figures, prev_selected_rows
    ):
        result = []
        plot_figures = plot_figures or {}
        selected_rows = selected_rows or {}
        prev_selected_rows = prev_selected_rows or {}
        alert_text = ""
        added_row = get_added_row(prev_selected_rows, selected_rows)
        spacer_hidden = len(data) > EXPERIMENTS_PAGE_SIZE
        if data and selected_rows:
            for row_num in selected_rows:
                alert_on_fail = row_num == added_row
                exp_id = data[row_num]["id"]
                try:
                    plots = _dashboard_data_reader.get_plot_data(exp_id)
                except EntropyError:
                    logger.exception(
                        f"Exception when getting plot data for exp_id={exp_id}"
                    )
                    if alert_on_fail:
                        alert_text = (
                            f"⚠ Error when reading plot data for this "
                            f"experiment. (id: {exp_id})"
                        )
                    plots = None
                if plots and len(plots) > 0:
                    failed_plot_ids = []
                    plot_figures = build_plot_tabs(
                        alert_on_fail, failed_plot_ids, plot_figures, plots, result
                    )
                    if len(failed_plot_ids) > 0:
                        alert_text = (
                            f"⚠ Some plots could not be rendered. "
                            f"(ids: {','.join(failed_plot_ids)})"
                        )
                else:
                    if alert_on_fail and alert_text == "":
                        alert_text = (
                            f"⚠ Experiment has no plots to render. (id: {exp_id})"
                        )
        if len(result) == 0:
            result = [build_plot_tabs_placeholder()]
        return result, plot_figures, selected_rows, alert_text, spacer_hidden

    def build_plot_tabs(alert_on_fail, failed_plot_ids, plot_figures, plots, result):
        for plot in plots:
            try:
                color = colors[len(result) % len(colors)]
                plot_tab, plot_figures = build_plot_tab_from_plot(
                    plot_figures, plot, color
                )
                result.append(plot_tab)
            except (EntropyError, TypeError):
                logger.exception(f"Failed to render plot id [{plot.id}]")
                if alert_on_fail:
                    plot_key = f"{plot.experiment_id}/{plot.id}"
                    failed_plot_ids.append(plot_key)
        return plot_figures

    def build_plot_tabs_placeholder():
        return dbc.Tab(
            html.Div(
                html.Div(
                    "Select an experiment above to display its plots here",
                    className="tab-placeholder-text",
                ),
                className="tab-placeholder-container",
            ),
            label="Plots",
            tab_id="plot-tab-placeholder",
        )

    def build_plot_tab_from_plot(
        plot_figures, plot: PlotRecord, color: str
    ) -> (dbc.Tab, Dict):
        plot_key = f"{plot.experiment_id}/{plot.id}"
        plot_name = f"Plot {plot_key}"
        plot_figure = go.Figure()
        plot.generator.plot_plotly(
            plot_figure,
            plot.plot_data,
            name=plot_name,
            color=color,
            showlegend=False,
        )
        plot_figure.update_layout(dark_plot_layout)
        plot_figures[plot_key] = dict(figure=plot_figure, color=color)
        return build_plot_tab(plot_figure, plot_name, plot_key), plot_figures

    def build_plot_tab(
        plot_figure: go.Figure, plot_name: str, plot_key: str
    ) -> dbc.Tab:
        return dbc.Tab(
            dcc.Graph(figure=plot_figure, responsive=True),
            label=plot_name,
            id=f"plot-tab-{plot_key}",
            tab_id=f"plot-tab-{plot_key}",
        )

    @_app.callback(
        Output("plot-keys-to-combine", "data"),
        Input("add-button", "n_clicks"),
        Input({"type": "remove-button", "index": ALL}, "n_clicks"),
        State("plot-tabs", "active_tab"),
        State("plot-keys-to-combine", "data"),
    )
    def add_or_remove_plot_keys_based_on_click_events(
        _, __, active_tab, plot_keys_to_combine
    ):
        plot_keys_to_combine = plot_keys_to_combine or []
        prop_id = dash.callback_context.triggered[0]["prop_id"]
        if prop_id == "add-button.n_clicks" and active_tab:
            active_plot_key = active_tab.replace("plot-tab-", "")
            if active_plot_key not in plot_keys_to_combine:
                plot_keys_to_combine.append(active_plot_key)
        elif "remove-button" in prop_id:
            id_dict = json.loads(prop_id.replace(".n_clicks", ""))
            remove_plot_key = id_dict["index"]
            if remove_plot_key in plot_keys_to_combine:
                plot_keys_to_combine.remove(remove_plot_key)
        return plot_keys_to_combine

    @_app.callback(
        Output("aggregate-tab", "children"),
        Output("remove-buttons", "children"),
        Input("plot-keys-to-combine", "data"),
        State("plot-figures", "data"),
    )
    def build_combined_plot_from_plot_keys(plot_keys_to_combine, plot_figures):
        if plot_keys_to_combine and len(plot_keys_to_combine) > 0:
            combined_figure = make_subplots(specs=[[{"secondary_y": True}]])
            remove_buttons = []
            for plot_id in plot_keys_to_combine:
                figure = plot_figures[plot_id]["figure"]
                color = plot_figures[plot_id]["color"]
                combined_figure.add_trace(figure["data"][0])
                button = build_remove_button(plot_id, color)
                remove_buttons.append(button)
            combined_figure.update_layout(dark_plot_layout)
            return (
                dcc.Graph(
                    id="aggregate-graph",
                    figure=combined_figure,
                    responsive=True,
                ),
                remove_buttons,
            )
        else:
            return (
                [build_aggregate_tab_placeholder()],
                [html.Div()],
            )

    def build_remove_button(plot_id, color):
        return dbc.Button(
            f"{plot_id} ✖️",
            style={"background-color": color},
            class_name="remove-button",
            id={"type": "remove-button", "index": plot_id},
        )

    def build_aggregate_tab_placeholder():
        return html.Div(
            [
                html.Div(
                    "Add a plot on the left to aggregate it here",
                    className="tab-placeholder-text",
                ),
                dcc.Graph(
                    id="aggregate-graph",
                    figure=go.Figure(),
                    responsive=True,
                    style={"display": "none"},
                ),
            ],
            className="tab-placeholder-container",
        )

    @_app.callback(
        Output("plot-tabs", "active_tab"),
        Input("plot-tabs", "children"),
    )
    def activate_last_plot_tab_when_tabs_are_changed(children):
        if len(children) > 0:
            last_tab = len(children) - 1
            return children[last_tab]["props"]["tab_id"]
        return 0

    @_app.callback(
        Output("aggregate-clipboard", "content"),
        Input("aggregate-clipboard", "n_clicks"),
        State("aggregate-graph", "figure"),
        prevent_initial_call=True,
    )
    def copy_aggregate_data_to_clipboard_as_python_code(_, figure):
        return _copy_aggregate_data_to_clipboard_as_python_code(_, figure)

    return _app


def get_added_row(prev: List[int], curr: List[int]) -> int or None:
    added_rows = list(set(curr) - set(prev))
    if len(added_rows) != 0:
        return added_rows[0]
    else:
        return None


def _copy_aggregate_data_to_clipboard_as_python_code(_, figure):
    _dict = list(map(lambda d: {d["name"]: {"x": d["x"], "y": d["y"]}}, figure["data"]))
    return f"data = {_dict}".replace("array(", "np.array(")



from dash import dash_table

EXPERIMENTS_PAGE_SIZE = 6


def table(records):
    tbl = dash_table.DataTable(
        id="experiments-table",
        columns=[
            dict(name="id", id="id", type="numeric"),
            dict(name="label", id="label", type="text"),
            dict(name="start_time", id="start_time", type="datetime"),
            dict(name="end_time", id="end_time", type="datetime"),
            dict(name="user", id="user", type="text"),
            dict(name="success", id="success"),
        ],
        data=records,
        persistence=True,
        persistence_type="session",
        row_selectable="multi",
        cell_selectable=False,
        sort_action="native",
        filter_action="native",
        page_action="native",
        page_size=EXPERIMENTS_PAGE_SIZE,
        style_header={"backgroundColor": "rgb(30, 30, 30)", "color": "white"},
        style_filter={"backgroundColor": "rgb(40, 40, 40)", "color": "white"},
        style_data={"backgroundColor": "rgb(50, 50, 50)", "color": "white"},
        style_cell={
            "textAlign": "left",
            "textOverflow": "ellipsis",
            "maxWidth": 0,
        },
        style_cell_conditional=[
            {"if": {"column_id": "id"}, "width": "7%"},
            {"if": {"column_id": "label"}, "width": "30%"},
            {
                "if": {"column_id": "start_time"},
                "width": "18%",
            },
            {"if": {"column_id": "end_time"}, "width": "18%"},
            {"if": {"column_id": "user"}, "width": "17%"},
            {"if": {"column_id": "success"}, "width": "10%"},
        ],
        style_data_conditional=[
            {
                "if": {
                    "column_id": "success",
                },
                "textAlign": "center",
            }
        ],
        # tooltip_data=[
        #     {
        #         column: {"value": str(value), "type": "markdown"}
        #         for column, value in row.items()
        #     }
        #     for row in records
        # ],
    )

    return tbl



from abc import ABC, abstractmethod
from dataclasses import dataclass
from datetime import datetime
from typing import List, Any, Optional, Iterable

from pandas import DataFrame

from entropylab.api.data_writer import PlotGenerator


class ScriptViewer:
    def __init__(self, stages: List[str]) -> None:
        super().__init__()
        self._stages = stages

    def show_stages(self):
        return "\n".join(self._stages)

    def print_stage(self, i):
        return self._stages[i]

    def print_all(self):
        return "\n".join(self._stages)


@dataclass
class ExperimentRecord:
    """
    Information about the experiment and the execution
    """

    id: int
    label: str
    script: ScriptViewer
    start_time: datetime
    end_time: Optional[datetime]
    story: str
    success: bool


@dataclass
class MetadataRecord:
    """
    A single record of metadata that was saved during the experiment
    """

    experiment_id: int
    id: str
    label: str
    stage: int
    data: Any
    time: datetime


@dataclass
class DebugRecord:
    """
    extra information about the experiment execution for better debugging
    """

    experiment_id: int
    id: int
    python_env: str
    python_history: str
    station_specs: str
    extra: str


@dataclass
class ResultRecord:
    """
    A single result that was saved during the experiment
    """

    experiment_id: int
    id: str
    label: str
    story: str
    stage: int
    data: Any
    time: datetime


@dataclass
class PlotRecord:
    """
    A single plot information and plotting instructions that was saved during the experiment
    """

    experiment_id: int
    id: int
    plot_data: Any = None
    generator: Optional[PlotGenerator] = None
    label: Optional[str] = None
    story: Optional[str] = None


@dataclass
class NodeResults:
    """
    All results of a specific node
    """

    execution_id: int
    results: Iterable[ResultRecord]


class DataReader(ABC):
    """
    An abstract class for Entropy database, defines the way entropy reads data.
    If a database implements both DataWriter and DataReader abstract classes,
    entropy will be able to read the results with a specific set of functions,
    and view them.
    """

    def __init__(self):
        super().__init__()

    def get_last_experiments(self, count: int) -> DataFrame:
        """
            read n number of last experiments to a pandas dataframe
        :param count: number of experiments
        """
        return self.get_experiments_range(0, count)

    @abstractmethod
    def get_experiments_range(self, starting_from_index: int, count: int) -> DataFrame:
        """
            read a range of experiments to a pandas dataframe
        :param starting_from_index: experiment index to start from (sorted desc by time)
        :param count: number of experiments
        """
        pass

    @abstractmethod
    def get_experiment_record(self, experiment_id: int) -> Optional[ExperimentRecord]:
        """
            returns a single experiment record, or None if experiment not found
        :param experiment_id: the id of experiment
        """
        pass

    @abstractmethod
    def get_experiments(
        self,
        label: Optional[str] = None,
        start_after: Optional[datetime] = None,
        end_after: Optional[datetime] = None,
        success: Optional[bool] = None,
    ) -> Iterable[ExperimentRecord]:
        """
            get multiple experiments records according to any combination of parameters filters
        :param label: experiment label to filter by
        :param start_after: experiments start after specific time
        :param end_after: experiments ended after specific time
        :param success: experiment success criteria
        """
        pass

    @abstractmethod
    def get_results(
        self,
        experiment_id: Optional[int] = None,
        label: Optional[str] = None,
        stage: Optional[int] = None,
    ) -> Iterable[ResultRecord]:
        """
            get multiple results according to any combination of parameters filters

        :param experiment_id: results from specific experiment
        :param label: results label to filter by
        :param stage: results stage within the experiment
        """
        pass

    @abstractmethod
    def get_metadata_records(
        self,
        experiment_id: Optional[int] = None,
        label: Optional[str] = None,
        stage: Optional[int] = None,
    ) -> Iterable[MetadataRecord]:
        """
            get multiple metadata records according to any combination of parameters filters

        :param experiment_id: metadata from specific experiment
        :param label: metadata label to filter by
        :param stage: metadata stage within the experiment
        """
        pass

    @abstractmethod
    def get_last_result_of_experiment(
        self, experiment_id: int
    ) -> Optional[ResultRecord]:
        """
        returns the last (by time) results of the requested experiment
        """
        pass

    @abstractmethod
    def get_debug_record(self, experiment_id: int) -> Optional[DebugRecord]:
        """
        returns debug information about the requested experiment
        """
        pass

    @abstractmethod
    def get_plots(self, experiment_id: int) -> List[PlotRecord]:
        """
        returns a list of all plots saved in the requested experiment
        """
        pass

    @abstractmethod
    def get_nodes_id_by_label(
        self, label: str, experiment_id: Optional[int] = None
    ) -> List[int]:
        """
            returns a list of all nodes with the given label
        :param label: results label
        :param experiment_id: optional to specify an experiment id, so
                only results from this experiment will be returned
        """
        pass

    def get_results_from_node(
        self,
        node_label: str,
        experiment_id: Optional[int] = None,
        result_label: Optional[str] = None,
    ) -> Iterable[NodeResults]:
        """
            returns multiple results with the given label, and combination of
            parameters filters

        :param node_label: the label of node
        :param experiment_id: optional to specify an experiment id, so
                only results from this experiment will be returned
        :param result_label: optional to specify a label of the results
        """
        nodes = self.get_nodes_id_by_label(node_label, experiment_id)
        if not nodes:
            raise KeyError(f"node {node_label} not found")
        nodes_results = []
        for node_id in nodes:
            nodes_results.append(
                NodeResults(
                    node_id,
                    self.get_results(
                        experiment_id,
                        label=result_label,
                        stage=node_id,
                    ),
                )
            )
        return nodes_results


class ExperimentReader:
    """
    Reads results and data from a single experiment
    """

    def __init__(self, experiment_id: int, db: DataReader) -> None:
        """
            Reads results and data from a single experiment
        :param experiment_id: the id of experiment
        :param db: results database that implemented the DataReader abstract class
        """
        super().__init__()
        self._experiment_id = experiment_id
        self._data_reader: DataReader = db

    def get_experiment_info(self) -> ExperimentRecord:
        """
        returns a record of the experiment info
        """
        return self._data_reader.get_experiment_record(self._experiment_id)

    def get_debug_record(self) -> Optional[DebugRecord]:
        """
        returns a record of extra debugging information about the experiment
        """
        return self._data_reader.get_debug_record(self._experiment_id)

    def get_metadata_records(
        self, label: Optional[str] = None
    ) -> Iterable[MetadataRecord]:
        """
            returns an iterable of records with the given label
        :param label: metadata label
        """
        return self._data_reader.get_metadata_records(self._experiment_id, label)

    def get_results(self, label: Optional[str] = None) -> Iterable[ResultRecord]:
        """
            returns an iterable of results record with the given label
        :param label: result label
        """
        return self._data_reader.get_results(self._experiment_id, label)

    def get_plots(self) -> List[PlotRecord]:
        """
        returns a list of plot records that were saved for current experiment
        """
        return self._data_reader.get_plots(self._experiment_id)



from typing import List, Dict

import dash_bootstrap_components as dbc
from dash import html, dcc

from entropylab.results.dashboard.table import table
from entropylab.results_backend.sqlalchemy.project import project_name, project_path


def layout(path: str, records: List[Dict]):
    return dbc.Container(
        className="main",
        children=[
            dcc.Store(id="plot-figures", storage_type="session"),
            dcc.Store(id="plot-keys-to-combine", storage_type="session"),
            dcc.Store(id="prev-selected-rows", storage_type="session"),
            dcc.Interval(
                id="interval", interval=3 * 1000, n_intervals=0  # in milliseconds
            ),
            dbc.Alert(
                "",
                id="failed-plotting-alert",
                color="warning",
                is_open=False,
                fade=True,
                # duration=3000,
            ),
            dbc.Modal(
                [
                    dbc.ModalHeader(
                        dbc.ModalTitle(
                            f"ℹ️ Project '{project_name(path)}' contains no experiments"
                        ),
                        close_button=False,
                    ),
                    dbc.ModalBody(
                        "As soon as experiments are saved to the project this notice "
                        "will be closed and the experiments will be shown below."
                    ),
                    dbc.ModalFooter(),
                ],
                id="empty-project-modal",
                backdrop="static",
                centered=True,
                is_open=False,
            ),
            dbc.Row(
                dbc.Navbar(
                    [
                        dbc.Col(
                            dbc.NavbarBrand(
                                html.A(
                                    html.Img(
                                        src="/assets/images/entropy_logo_dark.svg",
                                        width=150,
                                        id="entropy-logo",
                                    ),
                                    href="#",
                                ),
                                href="#",
                            ),
                            width="3",
                            id="logo-col",
                        ),
                        dbc.Col(
                            [
                                html.Div(f"{project_name(path)}", id="project-name"),
                                dbc.Tooltip(
                                    f"{project_path(path)}",
                                    target="project-name",
                                ),
                            ],
                            width="3",
                        ),
                        dbc.Col(
                            dbc.Row(
                                [
                                    dbc.Col(
                                        dbc.NavItem(
                                            dbc.NavLink(
                                                "Dashboard", href="#", active=True
                                            )
                                        )
                                    ),
                                    dbc.Col(
                                        dbc.NavItem(
                                            dbc.NavLink("Configuration", href="#")
                                        )
                                    ),
                                ]
                            ),
                            width="6",
                        ),
                    ],
                    color="primary",
                ),
            ),
            dbc.Row(
                dbc.Col(
                    [
                        html.H5("Experiments", id="experiments-title"),
                        (table(records)),
                        html.Div(id="no-paging-spacer"),
                    ],
                    width="12",
                )
            ),
            dbc.Row(
                [
                    dbc.Col(
                        (
                            dbc.Tabs(
                                id="plot-tabs",
                            ),
                        ),
                        width="5",
                    ),
                    dbc.Col(
                        html.Div(
                            dbc.Button(
                                "Add >>",
                                id="add-button",
                            ),
                            className="add-button-col-container",
                        ),
                        width="1",
                        className="add-button-col",
                    ),
                    dbc.Col(
                        [
                            dbc.Tabs(
                                id="aggregate-tabs",
                                children=[
                                    dbc.Tab(
                                        "",
                                        label="Aggregate",
                                        id="aggregate-tab",
                                    )
                                ],
                                persistence=True,
                            ),
                            dbc.Button(
                                [
                                    "Copy Data to Clipboard",
                                    dcc.Clipboard(
                                        title="Copy data to clipboard",
                                        id="aggregate-clipboard",
                                        className="position-absolute start-0 top-0 "
                                        "h-100 w-100 opacity-0",
                                    ),
                                ],
                                id="copy-data-button",
                                color="primary",
                            ),
                        ],
                        id="aggregate-container",
                        width="5",
                    ),
                    dbc.Col(
                        [
                            html.Div("<< Remove", id="remove-title"),
                            html.Div(id="remove-buttons"),
                        ],
                        width="1",
                    ),
                ]
            ),
        ],
    )



import pickle
from datetime import datetime
from enum import Enum
from typing import Optional, Any, Iterable, TypeVar, Callable

import h5py
import numpy as np

from entropylab import RawResultData
from entropylab.api.data_reader import ResultRecord, MetadataRecord
from entropylab.api.data_writer import Metadata
from entropylab.logger import logger
from entropylab.results_backend.sqlalchemy.model import (
    ResultDataType,
    ResultTable,
    Base,
)

T = TypeVar("T", bound=Base)
R = TypeVar("R", ResultRecord, MetadataRecord)


def _experiment_from(dset: h5py.Dataset) -> int:
    return dset.attrs["experiment_id"]


def _id_from(dset: h5py.Dataset) -> str:
    return dset.name


def _stage_from(dset: h5py.Dataset) -> int:
    return dset.attrs["stage"]


def _label_from(dset: h5py.Dataset) -> str:
    return dset.attrs["label"]


def _story_from(dset: h5py.Dataset) -> str:
    if "story" in dset.attrs:
        return dset.attrs["story"]
    return ""


def _data_from(dset: h5py.Dataset) -> Any:
    data = dset[()]
    if dset.dtype.metadata and dset.dtype.metadata.get("vlen") == str:
        return dset.asstr()[()]
    elif dset.attrs.get("data_type") == ResultDataType.Pickled.value:
        return pickle.loads(data)
    elif dset.attrs.get("data_type") == ResultDataType.String.value:
        # un-picklable data is stored as HDF5 Opaque so turn to bytes then to string:
        return data.tobytes().decode(encoding="utf-8")
    else:
        return data


def _time_from(dset: h5py.Dataset) -> datetime:
    return datetime.fromisoformat(dset.attrs["time"])


def _build_result_record(dset: h5py.Dataset) -> ResultRecord:
    return ResultRecord(
        experiment_id=_experiment_from(dset),
        id=_id_from(dset),
        label=_label_from(dset),
        story=_story_from(dset),
        stage=_stage_from(dset),
        data=_data_from(dset),
        time=_time_from(dset),
    )


def _build_metadata_record(dset: h5py.Dataset) -> MetadataRecord:
    return MetadataRecord(
        experiment_id=_experiment_from(dset),
        id=_id_from(dset),
        label=_label_from(dset),
        stage=_stage_from(dset),
        data=_data_from(dset),
        time=_time_from(dset),
    )


def _get_all_or_single(group: h5py.Group, name: Optional[str] = None):
    """
    Returns all or one child from an h5py.Group

    Parameters
    ----------
    group group to get child or children from. Can be h5py.File itself.
    name name of child to get. If None, indicates all children should be retrieved.

    Returns
    -------
    A list of group children (either h5py.Group or h5py.Datasets)
    """
    if name is None:
        return list(group.values())
    else:
        if str(name) in group:
            return [group[str(name)]]
        else:
            return []


class EntityType(Enum):
    RESULT = 1
    METADATA = 2


class _HDF5Reader:
    def _get_experiment_entities(
        self,
        entity_type: EntityType,
        convert_from_dset: Callable,
        experiment_id: Optional[int] = None,
        stage: Optional[int] = None,
        label: Optional[str] = None,
    ) -> Iterable[T]:
        dsets = []
        try:
            with self._open_hdf5("r") as file:
                if "experiments" in file:
                    top_group = file["experiments"]
                    exp_groups = _get_all_or_single(top_group, experiment_id)
                    for exp_group in exp_groups:
                        stage_groups = _get_all_or_single(exp_group, stage)
                        for stage_group in stage_groups:
                            label_groups = _get_all_or_single(stage_group, label)
                            for label_group in label_groups:
                                dset_name = entity_type.name.lower()
                                dset = label_group[dset_name]
                                dsets.append(convert_from_dset(dset))
            return dsets
        except FileNotFoundError:
            logger.exception("FileNotFoundError in get_experiment_entities()")
            return dsets

    def get_result_records(
        self,
        experiment_id: Optional[int] = None,
        stage: Optional[int] = None,
        label: Optional[str] = None,
    ) -> Iterable[ResultRecord]:
        entities = self._get_experiment_entities(
            EntityType.RESULT, _build_result_record, experiment_id, stage, label
        )
        return entities

    def get_metadata_records(
        self,
        experiment_id: Optional[int] = None,
        stage: Optional[int] = None,
        label: Optional[str] = None,
    ) -> Iterable[MetadataRecord]:
        entities = self._get_experiment_entities(
            EntityType.METADATA, _build_metadata_record, experiment_id, stage, label
        )
        return entities

    def get_last_result_of_experiment(
        self, experiment_id: int
    ) -> Optional[ResultRecord]:
        results = list(self.get_result_records(experiment_id, None, None))
        if results and len(results) > 0:
            results.sort(key=lambda x: x.time, reverse=True)
            return results[0]
        else:
            return None


class _HDF5Writer:
    def save_result(self, experiment_id: int, result: RawResultData) -> str:
        with self._open_hdf5("a") as file:
            return self._save_entity_to_file(
                file,
                EntityType.RESULT,
                experiment_id,
                result.stage,
                result.label,
                result.data,
                datetime.now(),
                result.story,
            )

    def save_metadata(self, experiment_id: int, metadata: Metadata):
        with self._open_hdf5("a") as file:
            return self._save_entity_to_file(
                file,
                EntityType.METADATA,
                experiment_id,
                metadata.stage,
                metadata.label,
                metadata.data,
                datetime.now(),
            )

    def _save_entity_to_file(
        self,
        file: h5py.File,
        entity_type: EntityType,
        experiment_id: int,
        stage: int,
        label: str,
        data: Any,
        time: datetime,
        story: Optional[str] = None,
        migrated_id: Optional[str] = None,
    ) -> str:
        path = f"/experiments/{experiment_id}/{stage}/{label}"
        group = file.require_group(path)
        dset = self._create_dataset(group, entity_type, data)
        dset.attrs.create("experiment_id", experiment_id)
        dset.attrs.create("stage", stage)
        dset.attrs.create("label", label)
        dset.attrs.create("time", time.astimezone().isoformat())
        if story:
            dset.attrs.create("story", story or "")
        if migrated_id:
            dset.attrs.create("migrated_id", migrated_id or "")
        return dset.name

    def _create_dataset(
        self, group: h5py.Group, entity_type: EntityType, data: Any
    ) -> h5py.Dataset:
        name = entity_type.name.lower()
        try:
            dset = group.create_dataset(name=name, data=data)
        except TypeError:
            data_type, pickled = self._pickle_data(data)
            # np.void turns our string to bytes (HDF5 Opaque):
            dset = group.create_dataset(name=name, data=np.void(pickled))
            dset.attrs.create("data_type", data_type.value, dtype="i2")
        return dset

    @staticmethod
    def _pickle_data(data: Any) -> (bytes, ResultDataType):
        try:
            pickled = pickle.dumps(data)
            data_type = ResultDataType.Pickled
        except Exception as ex:
            logger.debug("Could not pickle data, defaulting to __repr__()", ex)
            pickled = data.__repr__().encode(encoding="utf-8")
            data_type = ResultDataType.String
        return data_type, pickled


class _HDF5Migrator(_HDF5Writer):
    def migrate_result_rows(self, rows: Iterable[ResultTable]) -> None:
        self.migrate_rows(EntityType.RESULT, rows)

    def migrate_metadata_rows(self, rows: Iterable[Metadata]) -> None:
        self.migrate_rows(EntityType.METADATA, rows)

    def migrate_rows(self, entity_type: EntityType, rows: Iterable[T]) -> None:
        if rows is not None and len(list(rows)) > 0:
            with self._open_hdf5("a") as file:
                for row in rows:
                    if not row.saved_in_hdf5:
                        record = row.to_record()
                        hdf5_id = self._migrate_record(file, entity_type, record)
                        logger.debug(
                            f"Migrated ${entity_type.name} with id [{row.id}] "
                            f"to HDF5 with id [{hdf5_id}]"
                        )

    def _migrate_record(
        self, file: h5py.File, entity_type: EntityType, record: R
    ) -> str:
        return self._save_entity_to_file(
            file,
            entity_type,
            record.experiment_id,
            record.stage,
            record.label,
            record.data,
            record.time,
            getattr(record, "story", None),
            record.id,
        )

    def _migrate_result_record(
        self, file: h5py.File, result_record: ResultRecord
    ) -> str:
        return self._save_entity_to_file(
            file,
            EntityType.RESULT,
            result_record.experiment_id,
            result_record.stage,
            result_record.label,
            result_record.data,
            result_record.time,
            result_record.story,
            result_record.id,
        )


class HDF5Storage(_HDF5Reader, _HDF5Migrator, _HDF5Writer):
    def __init__(self, path=None):
        """Initializes a new or existing HDF5 file for storing experiment results
                 and metadata.

        :param path: filesystem path to the HDF5 file. If no path is given or the
                 path is empty, an in-memory-only HDF5 file is used.
        """
        if path is None or path == "":
            self._path = "./entropy.temp.hdf5"
            self._open_hdf5 = self._open_in_memory
        else:
            self._path = path
            self._open_hdf5 = self._open_in_fs
        self._check_file_permissions()

    def _check_file_permissions(self):
        file = self._open_hdf5("a")
        file.close()

    def _open_in_memory(self, mode: str) -> h5py.File:
        """Note that because backing_store=False, self._path is ignored & no file is
        saved on disk. See https://docs.h5py.org/en/stable/high/file.html#file-drivers
        """
        return h5py.File(self._path, mode, driver="core", backing_store=False)

    def _open_in_fs(self, mode: str) -> h5py.File:
        return h5py.File(self._path, mode)



import logging
import os
import sys
from typing import Optional

from waitress import serve

from entropylab.config import settings
from entropylab.results.dashboard.dashboard import build_dashboard_app
from entropylab.results.dashboard.theme import theme_stylesheet


def serve_dashboard(
    path: Optional[str] = None,
    host: Optional[str] = None,
    port: Optional[int] = None,
    debug: Optional[bool] = None,
):
    """Serves our "dashboard" app using waitress and opens it in a browser.

    :param path: The path to the Entropy project to connect to the dashboard.
    :param host: The host name from which to server the app. Defaults to "127.0.0.1".
    :param port: The port name from which to server the app. Defaults to 8050.
    :param debug: Start the dashboard in debug mode. Defaults to False.
    """
    if host is None:
        host = settings.get("dashboard.host", "127.0.0.1")
    if port is None:
        port = settings.get("dashboard.port", 8050)
    if debug is None:
        debug = settings.get("dashboard.debug", False)

    sys.path.append(os.path.abspath(path))
    app = build_dashboard_app(path)

    app.enable_dev_tools(debug=True)

    if debug:
        entropy_logger = logging.getLogger("entropy")
        entropy_logger.setLevel(logging.DEBUG)
        waitress_logger = logging.getLogger("waitress")
        waitress_logger.setLevel(logging.DEBUG)

    serve(app.server, host=host, port=port)



from typing import List

import plotly
import plotly.express as px
import plotly.graph_objects as go
from bokeh.models import Renderer
from bokeh.plotting import Figure
from matplotlib.axes import Axes as Axes
from matplotlib.figure import Figure as matplotlibFigure

from entropylab.api.data_writer import PlotGenerator


class LinePlotGenerator(PlotGenerator):
    def __init__(self) -> None:
        super().__init__()

    def plot_matplotlib(self, figure: matplotlibFigure, data, **kwargs) -> Renderer:
        raise NotImplementedError()

    def plot_bokeh(self, figure: Figure, data, **kwargs) -> Renderer:
        if isinstance(data, List) and len(data) == 2 and len(data[0]) == len(data[1]):
            x = data[0]
            y = data[1]
            return figure.line(
                x,
                y,
                color=kwargs.get("color", "blue"),
                legend_label=kwargs.get("label", ""),
            )
        else:
            raise TypeError("data type is not supported")

    def plot_plotly(self, figure: plotly.graph_objects.Figure, data, **kwargs):
        if isinstance(data, List) and len(data) == 2 and len(data[0]) == len(data[1]):
            x = data[0]
            y = data[1]
            color = kwargs.pop("color", "blue")
            figure.add_trace(
                go.Scatter(
                    mode="lines",
                    x=x,
                    y=y,
                    line_color=color,
                    **kwargs,
                )
            )
            return figure
        else:
            raise TypeError("data type is not supported")


class CirclePlotGenerator(PlotGenerator):
    def __init__(self) -> None:
        super().__init__()

    def plot_matplotlib(self, figure: Axes, data, **kwargs):
        raise NotImplementedError()

    def plot_bokeh(self, figure: Figure, data, **kwargs) -> Renderer:
        if isinstance(data, List) and len(data) == 2 and len(data[0]) == len(data[1]):
            x = data[0]
            y = data[1]
            color = (kwargs.get("color", "blue"),)
            return figure.circle(
                x,
                y,
                size=10,
                color=color,
                legend_label=kwargs.get("label", ""),
                alpha=0.5,
            )
        else:
            raise TypeError("data type is not supported")

    def plot_plotly(self, figure: plotly.graph_objects.Figure, data, **kwargs):
        if isinstance(data, List) and len(data) == 2 and len(data[0]) == len(data[1]):
            x = data[0]
            y = data[1]
            color = kwargs.pop("color", "blue")
            # noinspection PyTypeChecker
            figure.add_trace(
                go.Scatter(
                    mode="markers",
                    x=x,
                    y=y,
                    marker_color=color,
                    marker_size=10,
                    **kwargs,
                )
            )
            return figure
        else:
            raise TypeError("data type is not supported")


class ImShowPlotGenerator(PlotGenerator):
    def __init__(self) -> None:
        super().__init__()

    def plot_plotly(self, figure: go.Figure, data, **kwargs) -> None:
        headtmap_fig = px.imshow(data)
        figure.add_trace(headtmap_fig.data[0])

        return figure

    def plot_bokeh(self, figure: Figure, data, **kwargs) -> Renderer:
        raise NotImplementedError()

    def plot_matplotlib(self, figure: matplotlibFigure, data, **kwargs):
        raise NotImplementedError()



from entropylab.api.data_reader import ExperimentReader
from entropylab.api.data_writer import RawResultData
from entropylab.api.execution import EntropyContext
from entropylab.api.graph import GraphHelper
from entropylab.graph_experiment import (
    Graph,
    PyNode,
    SubGraphNode,
    pynode,
)
from entropylab.instruments.lab_topology import ExperimentResources, LabResources
from entropylab.results_backend.sqlalchemy.db import SqlAlchemyDB
from entropylab.script_experiment import Script, script_experiment

__all__ = [
    "ExperimentReader",
    "RawResultData",
    "EntropyContext",
    "GraphHelper",
    "Graph",
    "PyNode",
    "SubGraphNode",
    "pynode",
    "ExperimentResources",
    "LabResources",
    "SqlAlchemyDB",
    "Script",
    "script_experiment",
]



from __future__ import annotations

from abc import ABC, abstractmethod
from enum import Enum, unique
from typing import Dict, List, Optional

import pandas as pd


@unique
class MergeStrategy(Enum):
    OURS = 1
    THEIRS = 2


class ParamStore(ABC):
    @abstractmethod
    def to_dict(self):
        pass

    @abstractmethod
    def get(self, key, commit_id):
        """
            returns the value of a param by key

        :param key: the key identifying the param
        :param commit_id: an optional commit_id. if provided, the value will be
        returned from the specified commit
        """
        pass

    @abstractmethod
    def commit(self, label):
        pass

    @abstractmethod
    def checkout(self, commit_id: str, commit_num: int, move_by: int):
        pass

    @abstractmethod
    def list_commits(self, label: str):
        """
            returns a list of commits

        :param label: an optional label, if given then only commits that match
        it will be returned
        """
        pass

    @abstractmethod
    def list_values(self, key: str) -> pd.DataFrame:
        """
            list all the values of a given key taken from commit history,
            sorted by date ascending

        :param key: the key for which to list values
        :returns: a list of tuples where the values are, in order:
            - the value of the key
            - time of commit
            - commit_id
            - label assigned to commit
        """
        pass

    @abstractmethod
    def merge(
        self,
        theirs: Dict | ParamStore,
        merge_strategy: Optional[MergeStrategy] = MergeStrategy.OURS,
    ) -> None:
        pass

    """ Tags """

    @abstractmethod
    def add_tag(self, tag: str, key: str) -> None:
        pass

    @abstractmethod
    def remove_tag(self, tag: str, key: str) -> None:
        pass

    @abstractmethod
    def list_keys(self, tag: str) -> List[str]:
        pass

    """ Temporary State """

    @abstractmethod
    def save_temp(self) -> None:
        pass

    @abstractmethod
    def load_temp(self) -> None:
        pass

    @property
    @abstractmethod
    def is_dirty(self):
        """True iff params have been changed since the store has last been
        initialized or checked out"""
        pass



import enum
import importlib
import pickle
from datetime import datetime
from io import BytesIO
from typing import Any

import numpy as np
from sqlalchemy import (
    Column,
    Integer,
    String,
    DATETIME,
    ForeignKey,
    BLOB,
    Enum,
    Boolean,
)
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship

from entropylab.api.data_reader import (
    ExperimentRecord,
    ScriptViewer,
    ResultRecord,
    MetadataRecord,
    DebugRecord,
    PlotRecord,
)
from entropylab.api.data_writer import (
    ExperimentInitialData,
    RawResultData,
    Metadata,
    Debug,
    PlotSpec,
    NodeData,
)
from entropylab.api.errors import EntropyError
from entropylab.logger import logger


def _get_class(module_name, class_name):
    module = importlib.import_module(module_name)
    if not hasattr(module, class_name):
        raise EntropyError("class {} is not in {}".format(class_name, module_name))
    logger.debug("reading class {} from module {}".format(class_name, module_name))
    cls = getattr(module, class_name)
    return cls


def _encode_serialized_data(data):
    if isinstance(data, (np.ndarray, np.generic)):
        bio = BytesIO()
        np.save(bio, data)
        bio.seek(0)
        serialized_data = bio.read()
        data_type = ResultDataType.Npy
    else:
        try:
            serialized_data = pickle.dumps(data)
            data_type = ResultDataType.Pickled
        except RuntimeError:
            serialized_data = data.__repr__().encode(encoding="UTF-8")
            data_type = ResultDataType.String
    return data_type, serialized_data


def _decode_serialized_data(serialized_data, data_type):
    if data_type == ResultDataType.Pickled:
        data = pickle.loads(serialized_data)
    elif data_type == ResultDataType.Npy:
        bio = BytesIO(serialized_data)
        data = np.load(bio)
    else:
        data = serialized_data.decode()
    return data


Base = declarative_base()


class ExperimentTable(Base):
    __tablename__ = "Experiments"

    id = Column(Integer, primary_key=True)
    label = Column(String)
    script = Column(String)
    start_time = Column(DATETIME, nullable=False)
    end_time = Column(DATETIME)
    user = Column(String)
    story = Column(String)
    success = Column(Boolean, default=False)
    results = relationship("ResultTable", cascade="all, delete-orphan")
    experiment_metadata = relationship("MetadataTable", cascade="all, delete-orphan")
    debug = relationship("DebugTable", cascade="all, delete-orphan")
    plots = relationship("PlotTable", cascade="all, delete-orphan")

    def __repr__(self):
        return f"<_Experiment(id='{self.id}')>"

    def to_record(self):
        return ExperimentRecord(
            id=self.id,
            label=self.label,
            script=ScriptViewer([self.script]),
            start_time=self.start_time,
            end_time=self.end_time,
            story=self.story,
            success=self.success,
        )

    @staticmethod
    def from_initial_data(initial_data: ExperimentInitialData):
        return ExperimentTable(
            label=initial_data.label,
            script=initial_data.script,
            start_time=initial_data.start_time,
            user=initial_data.user,
            story=initial_data.story,
        )


class ResultDataType(enum.Enum):
    """Numeric values"""

    Pickled = 1
    String = 2
    Npy = 3


class ResultTable(Base):
    __tablename__: str = "Results"

    id = Column(Integer, primary_key=True)
    experiment_id = Column(Integer, ForeignKey("Experiments.id", ondelete="CASCADE"))
    stage = Column(Integer)
    story = Column(String)
    label = Column(String)
    time = Column(DATETIME, nullable=False)
    data = Column(BLOB)
    data_type = Column(Enum(ResultDataType))
    saved_in_hdf5 = Column(Boolean, nullable=False, default=False)

    def __repr__(self):
        return f"<Result(id='{self.id}')>"

    def to_record(self):
        data = _decode_serialized_data(self.data, self.data_type)
        return ResultRecord(
            experiment_id=self.experiment_id,
            id=self.id,
            label=self.label,
            story=self.story,
            stage=self.stage,
            data=data,
            time=self.time,
        )

    @staticmethod
    def from_model(experiment_id: int, result: RawResultData):
        data_type, serialized_data = _encode_serialized_data(result.data)
        return ResultTable(
            experiment_id=experiment_id,
            stage=result.stage,
            story=result.story,
            label=result.label,
            time=datetime.now(),
            data=serialized_data,
            data_type=data_type,
        )


class MetadataTable(Base):
    __tablename__ = "ExperimentMetadata"

    id = Column(Integer, primary_key=True)
    experiment_id = Column(Integer, ForeignKey("Experiments.id", ondelete="CASCADE"))
    stage = Column(Integer)
    label = Column(String)
    time = Column(DATETIME, nullable=False)
    data = Column(BLOB)
    data_type = Column(Enum(ResultDataType))
    saved_in_hdf5 = Column(Boolean, nullable=False, default=False)

    def __repr__(self):
        return f"<Metadata(id='{self.id}')>"

    def to_record(self):
        data = _decode_serialized_data(self.data, self.data_type)
        return MetadataRecord(
            experiment_id=self.experiment_id,
            id=self.id,
            label=self.label,
            stage=self.stage,
            data=data,
            time=self.time,
        )

    @staticmethod
    def from_model(experiment_id: int, metadata: Metadata):
        data_type, serialized_data = _encode_serialized_data(metadata.data)
        return MetadataTable(
            experiment_id=experiment_id,
            stage=metadata.stage,
            label=metadata.label,
            time=datetime.now(),
            data=serialized_data,
            data_type=data_type,
        )


class NodeTable(Base):
    __tablename__ = "Nodes"

    experiment_id = Column(
        Integer, ForeignKey("Experiments.id", ondelete="CASCADE"), primary_key=True
    )
    id = Column(Integer, primary_key=True)
    label = Column(String)
    start = Column(DATETIME, nullable=False)
    is_key_node = Column(Boolean)

    def __repr__(self):
        return f"<Node(exp_id='{self.experiment_id}', id='{self.id}')>"

    @staticmethod
    def from_model(experiment_id: int, node_data: NodeData):
        return NodeTable(
            experiment_id=experiment_id,
            id=node_data.node_id,
            start=node_data.start_time,
            label=node_data.label,
            is_key_node=node_data.is_key_node,
        )


class PlotTable(Base):
    __tablename__ = "Plots"

    id = Column(Integer, primary_key=True)
    experiment_id = Column(Integer, ForeignKey("Experiments.id", ondelete="CASCADE"))
    plot_data = Column(BLOB)
    data_type = Column(Enum(ResultDataType))
    generator_module = Column(String)
    generator_class = Column(String)
    time = Column(DATETIME)
    label = Column(String)
    story = Column(String)

    def __repr__(self):
        return f"<PlotSpec(id='{self.id}')>"

    def to_record(self) -> PlotRecord:
        data = _decode_serialized_data(self.plot_data, self.data_type)
        generator = _get_class(self.generator_module, self.generator_class)
        return PlotRecord(
            experiment_id=self.experiment_id,
            id=self.id,
            label=self.label,
            story=self.story,
            plot_data=data,
            generator=generator(),
        )

    @staticmethod
    def from_model(experiment_id: int, plot: PlotSpec, data: Any):
        data_type, serialized_data = _encode_serialized_data(data)
        return PlotTable(
            experiment_id=experiment_id,
            plot_data=serialized_data,
            data_type=data_type,
            generator_module=plot.generator.__module__,
            generator_class=plot.generator.__qualname__,
            time=datetime.now(),
            label=plot.label,
            story=plot.story,
        )


class DebugTable(Base):
    __tablename__ = "Debug"

    id = Column(Integer, primary_key=True)
    experiment_id = Column(Integer, ForeignKey("Experiments.id", ondelete="CASCADE"))
    python_env = Column(String)
    python_history = Column(String)
    station_specs = Column(String)
    extra = Column(String)

    def __repr__(self):
        return f"<Debug(id='{self.id}')>"

    def to_record(self):
        return DebugRecord(
            experiment_id=self.experiment_id,
            id=self.id,
            python_env=self.python_env,
            python_history=self.python_history,
            station_specs=self.station_specs,
            extra=self.extra,
        )

    @staticmethod
    def from_model(experiment_id: int, debug: Debug):
        return DebugTable(
            experiment_id=experiment_id,
            python_env=debug.python_env,
            python_history=debug.python_history,
            station_specs=debug.station_specs,
            extra=debug.extra,
        )



from pathlib import Path


""" Project related functions.

An Entropy *project* is a filesystem directory containing a '.entropy' directory
which itself contains an 'entropy.db' SQLite database file and an 'entropy.hdf5'
HDF5 file.

project
    .entropy
        entropy.db
        entropy.hdf5
"""


def project_name(path: str) -> str:
    """Returns a project name for the given path

    :path: A filesystem path to a directory. Directory does not have to contain an
    actual Entropy project"""
    return Path(path).absolute().name


def project_path(path: str) -> str:
    """Returns the absolute path for the given path, as a string

    :path: A filesystem path to a directory. Directory does not have to contain an
    actual Entropy project"""
    return Path(path).absolute()



import dash_bootstrap_components as dbc

theme_stylesheet = dbc.themes.DARKLY

# Based on https://bootswatch.com/darkly/
colors = [
    "#375a7f",
    "#3a3a3a",
    "#00bc8c",
    "#f39c12",
    "#e74c3c",
    "#3498db",
    "#adb5bd",
]
plot_font_color = "#fff"
plot_legend_font_color = "#fff"
plot_paper_bgcolor = "rgb(48, 48, 48)"
plot_plot_bgcolor = "rgb(173, 181, 189)"

dark_plot_layout = dict(
    font_color=plot_font_color,
    legend_font_color=plot_legend_font_color,
    paper_bgcolor=plot_paper_bgcolor,
    plot_bgcolor=plot_plot_bgcolor,
    showlegend=False,
    margin=dict(b=40, t=40, l=40, r=40),
)



import logging
from datetime import datetime
from typing import List, TypeVar, Optional, ContextManager, Iterable, Union, Any
from typing import Set

import jsonpickle
import pandas as pd
from pandas import DataFrame
from sqlalchemy import desc
from sqlalchemy.exc import DBAPIError
from sqlalchemy.orm import sessionmaker, Session
from sqlalchemy.sql import Selectable
from sqlalchemy.util.compat import contextmanager

from entropylab.api.data_reader import (
    DataReader,
    ExperimentRecord,
    ResultRecord,
    MetadataRecord,
    DebugRecord,
    PlotRecord,
)
from entropylab.api.data_writer import (
    DataWriter,
    ExperimentInitialData,
    ExperimentEndData,
    RawResultData,
    Metadata,
    Debug,
    PlotSpec,
    NodeData,
)
from entropylab.config import settings
from entropylab.instruments.instrument_driver import Function, Parameter
from entropylab.instruments.lab_topology import (
    PersistentLabDB,
    DriverType,
    ResourceRecord,
)
from entropylab.results_backend.sqlalchemy.db_initializer import _DbInitializer
from entropylab.results_backend.sqlalchemy.lab_model import (
    Resources,
    ResourcesSnapshots,
)
from entropylab.results_backend.sqlalchemy.model import (
    ExperimentTable,
    PlotTable,
    ResultTable,
    DebugTable,
    MetadataTable,
    NodeTable,
)

T = TypeVar(
    "T",
)


class SqlAlchemyDB(DataWriter, DataReader, PersistentLabDB):
    """
    Database implementation using SqlAlchemy package for results (DataWriter
     and DataReader) and lab resources (PersistentLabDB)
    """

    def __init__(self, path=None, echo=False, **kwargs):
        """
            Initializes database and HDF5 files for an Entropy project
        :param path: path to directory containing an Entropy project
        :param echo: if True, the database engine will log all statements
        """
        super(SqlAlchemyDB, self).__init__()
        self._enable_hdf5_storage = kwargs.get("enable_hdf5_storage")
        self._engine, self._storage = _DbInitializer(path, echo=echo).init_db()
        self._Session = sessionmaker(bind=self._engine)

    def save_experiment_initial_data(self, initial_data: ExperimentInitialData) -> int:
        transaction = ExperimentTable.from_initial_data(initial_data)
        return self._execute_transaction(transaction)

    def save_experiment_end_data(self, experiment_id: int, end_data: ExperimentEndData):
        with self._session_maker() as sess:
            query = (
                sess.query(ExperimentTable)
                .filter(ExperimentTable.id == int(experiment_id))
                .one_or_none()
            )
            if query:
                query.end_time = end_data.end_time
                query.success = end_data.success
                sess.flush()

    def save_result(self, experiment_id: int, result: RawResultData):
        if result.label is None:
            raise TypeError("result.label cannot be None")
        if result.label == "":
            raise ValueError("result.label cannot be empty")
        saved_in_hdf5 = False
        if self.__hdf5_storage_enabled():
            try:
                self._storage.save_result(experiment_id, result)
                saved_in_hdf5 = True
            except ValueError as ex:
                raise ValueError(
                    f"Result already exists (experiment_id='{experiment_id}', "
                    f"stage='{result.stage}', label='{result.label}'"
                ) from ex
            except RuntimeError:
                logging.exception("Error saving result to HDF5")
                saved_in_hdf5 = False
        transaction = ResultTable.from_model(experiment_id, result)
        transaction.saved_in_hdf5 = saved_in_hdf5
        return self._execute_transaction(transaction)

    def save_metadata(self, experiment_id: int, metadata: Metadata):
        if metadata.label is None:
            raise TypeError("metadata.label cannot be None")
        if metadata.label == "":
            raise ValueError("metadata.label cannot be empty")
        saved_in_hdf5 = False
        if self.__hdf5_storage_enabled():
            try:
                self._storage.save_metadata(experiment_id, metadata)
                saved_in_hdf5 = True
            except ValueError as ex:
                raise ValueError(
                    f"Metadata already exists (experiment_id='{experiment_id}', "
                    f"stage='{metadata.stage}', label='{metadata.label}'"
                ) from ex
            except RuntimeError:
                logging.exception("Error saving metadata to HDF5.")
                saved_in_hdf5 = False
        transaction = MetadataTable.from_model(experiment_id, metadata)
        transaction.saved_in_hdf5 = saved_in_hdf5
        return self._execute_transaction(transaction)

    def save_debug(self, experiment_id: int, debug: Debug):
        transaction = DebugTable.from_model(experiment_id, debug)
        return self._execute_transaction(transaction)

    def save_plot(self, experiment_id: int, plot: PlotSpec, data: Any):
        transaction = PlotTable.from_model(experiment_id, plot, data)
        return self._execute_transaction(transaction)

    def save_node(self, experiment_id: int, node_data: NodeData):
        transaction = NodeTable.from_model(experiment_id, node_data)
        return self._execute_transaction(transaction)

    def get_experiments_range(self, starting_from_index: int, count: int) -> DataFrame:
        with self._session_maker() as sess:
            query = (
                sess.query(ExperimentTable)
                .with_entities(
                    ExperimentTable.id,
                    ExperimentTable.label,
                    ExperimentTable.start_time,
                    ExperimentTable.end_time,
                    ExperimentTable.user,
                    ExperimentTable.success,
                )
                .slice(starting_from_index, starting_from_index + count)
            )
            return self._query_pandas(query)

    def get_experiment_record(self, experiment_id: int) -> Optional[ExperimentRecord]:
        with self._session_maker() as sess:
            query = (
                sess.query(ExperimentTable)
                .filter(ExperimentTable.id == int(experiment_id))
                .one_or_none()
            )
            if query:
                return query.to_record()

    def get_experiments(
        self,
        label: Optional[str] = None,
        start_after: Optional[datetime] = None,
        end_after: Optional[datetime] = None,
        success: Optional[bool] = None,
    ) -> Iterable[ExperimentRecord]:
        with self._session_maker() as sess:
            query = sess.query(ExperimentTable)
            if label is not None:
                query = query.filter(ExperimentTable.label == label)
            if success is not None:
                query = query.filter(ExperimentTable.success == success)
            if start_after is not None:
                query = query.filter(ExperimentTable.start_time > start_after)
            if end_after is not None:
                query = query.filter(ExperimentTable.end_time > end_after)
            return [item.to_record() for item in query.all()]

    def get_results(
        self,
        experiment_id: Optional[int] = None,
        label: Optional[str] = None,
        stage: Optional[int] = None,
    ) -> Iterable[ResultRecord]:
        if self.__hdf5_storage_enabled():
            return self._storage.get_result_records(experiment_id, stage, label)
        else:
            return self.__get_results_from_sqlalchemy(experiment_id, label, stage)

        pass

    def __get_results_from_sqlalchemy(
        self,
        experiment_id: Optional[int] = None,
        label: Optional[str] = None,
        stage: Optional[int] = None,
        saved_in_hdf5: Optional[bool] = None,
    ) -> Iterable[ResultRecord]:
        with self._session_maker() as sess:
            query = sess.query(ResultTable)
            if experiment_id is not None:
                query = query.filter(ResultTable.experiment_id == int(experiment_id))
            if label is not None:
                query = query.filter(ResultTable.label == str(label))
            if stage is not None:
                query = query.filter(ResultTable.stage == int(stage))
            if self.__hdf5_storage_enabled() and saved_in_hdf5 is not None:
                query = query.filter(ResultTable.saved_in_hdf5 == bool(saved_in_hdf5))
            return [item.to_record() for item in query.all()]

    def get_metadata_records(
        self,
        experiment_id: Optional[int] = None,
        label: Optional[str] = None,
        stage: Optional[int] = None,
    ) -> Iterable[MetadataRecord]:
        with self._session_maker() as sess:
            query = sess.query(MetadataTable)
            if experiment_id is not None:
                query = query.filter(MetadataTable.experiment_id == int(experiment_id))
            if label is not None:
                query = query.filter(MetadataTable.label == label)
            if stage is not None:
                query = query.filter(MetadataTable.stage == stage)
            return [item.to_record() for item in query.all()]

    def get_debug_record(self, experiment_id: int) -> Optional[DebugRecord]:
        with self._session_maker() as sess:
            query = (
                sess.query(MetadataTable)
                .filter(DebugTable.experiment_id == int(experiment_id))
                .one_or_none()
            )
            if query:
                return query.to_record()

    def get_all_results_with_label(self, exp_id, name) -> DataFrame:
        with self._session_maker() as sess:
            query = (
                sess.query(ResultTable)
                .filter(ResultTable.experiment_id == int(exp_id))
                .filter(ResultTable.label == name)
            )
            return self._query_pandas(query)

    def get_plots(self, experiment_id: int) -> List[PlotRecord]:
        with self._session_maker() as sess:
            query = (
                sess.query(PlotTable)
                .filter(PlotTable.experiment_id == int(experiment_id))
                .all()
            )
            if query:
                return [plot.to_record() for plot in query]
        return []

    def get_nodes_id_by_label(
        self, label: str, experiment_id: Optional[int] = None
    ) -> List[int]:
        with self._session_maker() as sess:
            query = sess.query(NodeTable).filter(NodeTable.label == label)
            if experiment_id is not None:
                query = query.filter(NodeTable.experiment_id == int(experiment_id))

            result = query.all()
            if result:
                return [node.id for node in result]
        return []

    def get_last_result_of_experiment(
        self, experiment_id: int
    ) -> Optional[ResultRecord]:
        if self.__hdf5_storage_enabled():
            return self._storage.get_last_result_of_experiment(experiment_id)
        else:
            return self.__get_last_result_of_experiment_from_sqlalchemy(experiment_id)

    def __get_last_result_of_experiment_from_sqlalchemy(
        self, experiment_id: int
    ) -> Optional[ResultRecord]:
        with self._session_maker() as sess:
            query = (
                sess.query(ResultTable)
                .filter(ResultTable.experiment_id == int(experiment_id))
                .order_by(desc(ResultTable.time))
                .first()
            )
            if query:
                return query.to_record()

    def custom_query(self, query: Union[str, Selectable]) -> DataFrame:
        with self._session_maker() as sess:
            if isinstance(query, str):
                selectable = query
            else:
                selectable = query.statement

            return pd.read_sql(selectable, sess.bind)

    def _execute_transaction(self, transaction):
        with self._session_maker() as sess:
            sess.add(transaction)
            sess.flush()
            return transaction.id

    @staticmethod
    def _query_pandas(query):
        return pd.read_sql(query.statement, query.session.bind)

    @contextmanager
    def _session_maker(self) -> ContextManager[Session]:
        """Provide a transactional scope around a series of operations."""
        session = self._Session()
        try:
            yield session
            session.commit()
        except DBAPIError:
            session.rollback()
            raise
        finally:
            session.close()

    def save_new_resource_driver(
        self,
        name: str,
        driver_source_code: str,
        module: str,
        class_name: str,
        serialized_args: str,
        serialized_kwargs: str,
        number_of_experiment_args: int,
        keys_of_experiment_kwargs: List[str],
        functions: List[Function],
        parameters: List[Parameter],
        undeclared_functions: List[Function],
    ):
        transaction = Resources(
            update_time=datetime.now(),
            name=name,
            driver=driver_source_code,
            module=module,
            class_name=class_name,
            version=1,
            driver_type=DriverType.Packaged,
            args=serialized_args,
            kwargs=serialized_kwargs,
            number_of_experiment_args=number_of_experiment_args,
            keys_of_experiment_kwargs=", ".join(keys_of_experiment_kwargs),
            cached_metadata=jsonpickle.dumps(
                {
                    "functions": functions,
                    "undeclared_functions": undeclared_functions,
                    "parameters": parameters,
                }
            ),
        )
        return self._execute_transaction(transaction)

    def remove_resource(self, name: str):
        with self._session_maker() as sess:
            query = (
                sess.query(Resources)
                .filter(Resources.name == name)
                .order_by(desc(Resources.update_time))
                .first()
            )
            if query:
                query.deleted = True
                sess.flush()

    def save_state(self, name: str, state: str, snapshot_name: str):
        driver_id = self._get_driver_id(name)
        transaction = ResourcesSnapshots(
            update_time=datetime.now(),
            name=snapshot_name,
            driver_id=driver_id,
            state=state,
        )
        return self._execute_transaction(transaction)

    def get_state(self, resource_name: str, snapshot_name: str) -> str:
        driver_id = self._get_driver_id(resource_name)
        with self._session_maker() as sess:
            query = (
                sess.query(ResourcesSnapshots)
                .filter(ResourcesSnapshots.driver_id == driver_id)
                .filter(ResourcesSnapshots.name == snapshot_name)
                .order_by(desc(ResourcesSnapshots.update_time))
                .first()
            )
            if query:
                return query.state
            else:
                return ""

    def get_all_states(self, name) -> Iterable[str]:
        driver_id = self._get_driver_id(name)
        with self._session_maker() as sess:
            query = (
                sess.query(ResourcesSnapshots)
                .filter(ResourcesSnapshots.driver_id == driver_id)
                .order_by(desc(ResourcesSnapshots.update_time))
            )
            return query.all()

    def get_resource(self, name) -> Optional[ResourceRecord]:
        with self._session_maker() as sess:
            query = (
                sess.query(Resources)
                .filter(Resources.name == name)
                .order_by(desc(Resources.update_time))
                .first()
            )
            if query and not query.deleted:
                cached_metadata = jsonpickle.loads(query.cached_metadata)
                return ResourceRecord(
                    query.name,
                    query.module,
                    query.class_name,
                    query.driver,
                    query.version,
                    query.driver_type,
                    query.args,
                    query.kwargs,
                    cached_metadata["functions"],
                    cached_metadata["parameters"],
                    cached_metadata["undeclared_functions"],
                )
            else:
                return None

    def _get_driver_id(self, name) -> int:
        with self._session_maker() as sess:
            query = (
                sess.query(Resources)
                .filter(Resources.name == name)
                .order_by(desc(Resources.update_time))
                .first()
            )
            if query and not query.deleted:
                return int(query.id)
            else:
                return -1

    def set_locked(self, resource_name):
        pass

    def set_released(self, resource_name):
        pass

    def get_all_resources(self) -> Set[str]:
        with self._session_maker() as sess:
            query = (
                sess.query(Resources)
                .filter(Resources.deleted == False)  # noqa: E712
                .group_by(Resources.name)
                .all()
            )
            if query is not None:
                return {item.name for item in query}
            else:
                return set()

    def __hdf5_storage_enabled(self) -> bool:
        """Feature toggle for 'hdf5 storage' feature

        Class member set in __init__() overrides config setting"""
        if self._enable_hdf5_storage is None:
            enabled = settings.get("toggles.hdf5_storage", True)
        else:
            enabled = self._enable_hdf5_storage
        return enabled



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['entropylab',
 'entropylab.api',
 'entropylab.cli',
 'entropylab.instruments',
 'entropylab.results.dashboard',
 'entropylab.results_backend',
 'entropylab.results_backend.sqlalchemy',
 'entropylab.results_backend.sqlalchemy.alembic',
 'entropylab.results_backend.sqlalchemy.alembic.versions']

package_data = \
{'': ['*'], 'entropylab.results.dashboard': ['assets/*', 'assets/images/*']}

install_requires = \
['alembic>=1.6.5,<2.0.0',
 'bokeh>=2.3.0,<3.0.0',
 'dash-bootstrap-components>=1.0.0,<2.0.0',
 'dash>=2.1.0,<3.0.0',
 'dill>=0.3.3,<0.4.0',
 'dynaconf>=3.1.4,<4.0.0',
 'graphviz>=0.16,<0.17',
 'h5py>=3.3.0,<4.0.0',
 'jsonpickle>=2.0.0,<3.0.0',
 'matplotlib>=3.4.1,<4.0.0',
 'munch>=2.5.0,<3.0.0',
 'networkx>=2.5.1,<3.0.0',
 'numpy>=1.20.1,<2.0.0',
 'pandas>=1.2.3,<2.0.0',
 'param>=1.10.1,<2.0.0',
 'sqlalchemy>=1.4.0,<2.0.0',
 'tinydb>=4.5.2,<5.0.0',
 'waitress>=2.0.0,<3.0.0']

entry_points = \
{'console_scripts': ['entropy = entropylab.cli.main:main',
                     'n3p = entropylab.cli.main:main']}

setup_kwargs = {
    'name': 'entropylab',
    'version': '0.5.0',
    'description': '',
    'long_description': '![PyPI](https://img.shields.io/pypi/v/entropylab)\n[![discord](https://img.shields.io/discord/806244683403100171?label=QUA&logo=Discord&style=plastic)](https://discord.gg/7FfhhpswbP)\n\n[![License](https://img.shields.io/badge/License-BSD%203--Clause-blue.svg)](https://opensource.org/licenses/BSD-3-Clause)\n[![DOI](https://zenodo.org/badge/DOI/10.5281/zenodo.4785097.svg)](https://doi.org/10.5281/zenodo.4785097)\n\n# Entropy\n\nEntropy is a lab workflow managment package built for, but not limitied-to, streamlining the process of running quantum information processing experiments. \n\nEntropy is built to solve a few major hurdles in experiment design: \n\n1. Building, maintaining and executing complex experiments\n2. Data collection\n3. Device management\n4. Calibration automation\n\nTo tackle these problems, Entropy is built around the central concept of a graph strucutre. The nodes of a graph give us a convenient way \nto brake down experiments into stages and to automate some of the tasks required in each node. For example data collection is automated, at least in part, \nby saving node data and code to a persistant database. \n\nDevice managment is the challange of managing the state and control of a variety of different resources. These include, but are not limited to, lab instrumnets. \nThey can also be computational resources, software resources or others. Entropy is built with tools to save such resources to a shared database and give nodes access to \nthe resources needed during an experiment. \n\nPerforming automatic calibration is an important reason why we built Entropy. This could be though of as the usecase most clearly benefiting from shared resources, persistant \nstorage of different pieced of information and the graph structure. If the final node in a graph is the target experiment, then all the nodes between the root and that node are often \ncalibration steps. The documentation section will show how this can be done. \n\nThe Entropy system is built with concrete implemnetations of the various parts (database backend, resource managment and others) but is meant to be completely customizable. Any or every part of the system can be tailored by end users. \n\n## Versioning and the Alpha release \n\nThe current release of Entropy is version 0.1.0. You can learn more about the Entropy versioning scheme in the versioning\ndocument. This means this version is a work in progress in several important ways: \n\n1. It is not fully tested\n2. There are important features missing, such as the results GUI which will enable visual results viewing and automatic plotting\n3. There will more than likely be breaking changes to the API for a while until we learn how things should be done. \n\nKeep this in mind as you start your journey. \n\n## Installation\n\nInstallation is done from pypi using the following command\n\n```shell\npip install entropylab\n```\n\n## Testing your installation\n\nimport the library from `entropylab`\n\n```python\nfrom entropylab import *\n\ndef my_func():\n    return {\'res\': 1}\n\nnode1 = PyNode("first_node", my_func, output_vars={\'res\'})\nexperiment = Graph(None, {node1}, "run_a")  # No resources used here\nhandle = experiment.run()\n```\n\n## Usage\n\nSee [docs](docs) folder in this repository for all the dirty details.\n\n\n## Extensions\n\nEntropy can and will be extended via custom extensions. An example is `entropylab-qpudb`, an extension built to keep track of the calibration parameters of a mutli-qubit Quantum Processing Unit (QPU). This extension is useful when writing an automatic calibration graph. \n\n\n\n',
    'author': 'Guy Kerem',
    'author_email': 'guy@quantum-machines.co',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/entropy-lab/entropy',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7.1,<4.0.0',
}


setup(**setup_kwargs)



from typing import Type, Union, Optional, List

import jsonpickle
from qcodes import Function, ChannelList
from qcodes.instrument.base import InstrumentBase
from qcodes.instrument.parameter import _BaseParameter

from entropylab.api.errors import EntropyError
from entropylab.instruments.instrument_driver import (
    Parameter,
    Function as EntropyFunction,
    Entropy_Resource_Name,
    DriverSpec,
    Resource,
)
from entropylab.logger import logger


def _transform_parameter(parameter: _BaseParameter) -> Parameter:
    if hasattr(parameter, "unit"):
        unit = parameter.unit
    else:
        unit = None
    return Parameter(
        parameter.name,
        unit,
        parameter.step,
        parameter.scale,
        parameter.offset,
        parameter.vals,
        "",
    )


def _transform_function(function: Function) -> EntropyFunction:
    return EntropyFunction(function.name)


def _get_transformed_parameters(parameters):
    return [_transform_parameter(p) for p in parameters]


def _get_transformed_functions(functions):
    return [_transform_function(f) for f in functions]


def _get_undeclared_functions(instance, functions: List[EntropyFunction]):
    existing_functions_names = [func.name for func in functions]
    instrument_base_functions = [
        func
        for func in dir(InstrumentBase)
        if callable(getattr(InstrumentBase, func)) and not func.startswith("__")
    ]
    method_list = [
        func
        for func in dir(instance)
        if callable(getattr(instance, func))
        and not func.startswith("__")
        and func not in instrument_base_functions
        and not isinstance(getattr(instance, func), _BaseParameter)
        and func not in existing_functions_names
    ]
    entropy_extra_methods = [EntropyFunction(func) for func in method_list]
    return entropy_extra_methods


class QcodesAdapter(Resource):
    """"""

    def __init__(self, driver: Type[InstrumentBase], *args, **kwargs):
        super().__init__(**kwargs)
        self._instance: InstrumentBase = None
        self._driver = driver
        self._args = args
        self._kwargs = kwargs
        self._kwargs.pop(Entropy_Resource_Name, None)
        self._is_connected = False

    def get_dynamic_driver_specs(self) -> DriverSpec:
        """
        qcodes drivers only get a complete list of parameters and actions when loading
        and connecting to device
        this function will add all the parameters to the driver, can be used outside and saved

        """
        self.connect()
        parameters = _get_transformed_parameters(self._instance.parameters.values())
        functions = _get_transformed_functions(self._instance.functions.values())
        undeclared_functions = _get_undeclared_functions(self._instance, functions)
        for name, sub in self._instance.submodules:
            sub_params, sub_functions = self._extract_submodule_specs(name, sub)
            parameters.extend(sub_params)
            functions.extend(sub_functions)

        self.teardown()
        return DriverSpec(parameters, functions, undeclared_functions)

    def _extract_submodule_specs(
        self, submodule_name, submodule: Union["InstrumentBase", "ChannelList"]
    ):
        parameters = _get_transformed_parameters(submodule.parameters)
        functions = _get_transformed_functions(submodule.functions)

        for p in parameters:
            p.submodule = f"{submodule_name}"
        for f in functions:
            f.submodule = f"{submodule_name}"

        for name, sub in self._instance.submodules:
            self._extract_submodule_specs(f"{submodule_name}_{name}", sub)

        return parameters, functions

    def connect(self):
        self._is_connected = True
        self._instance = self._driver(*self._args, **self._kwargs)

    def get_instance(self):
        if self._is_connected:
            return self._instance
        else:
            self.connect()
            return self._instance

    def teardown(self):
        try:
            if self._is_connected:
                del self._instance
            self._is_connected = False
        except Exception as e:
            logger.debug("Exception during resource teardown", e)

    @property
    def instance(self):
        return self._instance

    def snapshot(self, update: Optional[bool] = True) -> str:
        """
        Use qcodes snapshot method to save the state of the instrument.
        the qcodes snapshot dictionary is then serialized using jsonpickle,
        or using the str() function if dictionary can not be pickled.
        """
        if self._is_connected:
            qcodes_snapshot_dict = self._instance.snapshot(update=update)
            try:
                serialized_snapshot = jsonpickle.dumps(qcodes_snapshot_dict)
            except Exception:
                serialized_snapshot = str(qcodes_snapshot_dict)
            return serialized_snapshot
        else:
            raise EntropyError("can not save snapshot of disconnected instrument")

    def revert_to_snapshot(self, snapshot: str):
        raise NotImplementedError(
            f"resource {self.__class__.__qualname__} has not implemented revert to snapshot"
        )
